import numpy as np
from scipy.optimize import minimize

from .PointHomogeneous import PointHomogeneous


class MiniBall:
    def __init__(self,
                 points: list[PointHomogeneous],
                 metric: "AffineMetric" = None):
        """
        Initialize the MiniBall class

        :param list[PointHomogeneous] points: array of points in the space
        :param AffineMetric metric: alternative metric to be used for the ball
        """
        self.points = points

        self.number_of_points = len(self.points)
        self.dimension = self.points[0].coordinates.size

        self.metric = metric

        if metric is None:
            self.metric = 'euclidean'
            self.metric_obj = None
        else:
            metric = 'hofer'
            self.metric_obj = metric

        self.center = np.zeros(self.dimension)
        self.radius = 10.0

        self.optimization_results = self.get_ball()
        self.center = PointHomogeneous(self.optimization_results.x[:-1])
        self.radius = self.optimization_results.x[-1]

    def get_ball(self):
        """
        Find the smallest ball containing all given points in Euclidean metric
        """

        def objective_function(x):
            """
            Objective function to minimize the squared radius r^2 of the ball
            """
            return x[-1] ** 2

        # Prepare constraint equations based on the metric
        if self.metric == "hofer":
            def constraint_equations(x):
                """
                For Hofer metric, constraint equations must satisfy the ball by:
                r - radius of the sphere, x - one of given points,
                c - center of the sphere
                """
                constraints = np.zeros(self.number_of_points)

                for i in range(self.number_of_points):
                    squared_distance = sum(self.metric_obj.squared_distance_pr12_points(
                        self.points[i].normalize(), x[j])
                        for j in range(self.dimension)
                    )
                    constraints[i] = x[-1] ** 2 - squared_distance
                return constraints
        else:
            def constraint_equations(x):
                """
                For Euclidean metric, constraint equations must satisfy the ball by:
                r^2 - (x - c)^2 >= 0
                r - radius of the sphere, x - one of given points,
                c - center of the sphere
                """
                constraints = np.zeros(self.number_of_points)
                for i in range(self.number_of_points):
                    # in case of Euclidean metric, the normalized point has to be taken
                    # in account
                    squared_distance = sum(
                        (self.points[i].normalize()[j] - x[j]) ** 2
                        for j in range(self.dimension)
                    )
                    constraints[i] = x[-1] ** 2 - squared_distance
                return constraints

        # Prepare inequality constraint dictionary
        ineq_con = {"type": "ineq", "fun": constraint_equations}

        # Initialize optimization variables
        initial_guess = np.zeros(self.dimension + 1)
        initial_guess[0] = 1.0
        initial_guess[-1] = self.radius

        # Perform optimization
        result = minimize(objective_function, initial_guess, constraints=ineq_con)
        return result

    def get_plot_data(self) -> tuple:
        """
        Get data for plotting in 3D space

        :return: x, y, z coordinates of the ball surface
        :rtype: tuple

        :raises ValueError: if the dimension is not 4 or 13
        """
        if self.dimension == 4 or self.dimension == 13:
            # Create the 3D sphere representing the circle
            u = np.linspace(0, 2 * np.pi, 30)
            v = np.linspace(0, np.pi, 30)

            x = (self.radius * np.outer(np.cos(u), np.sin(v))
                 + self.center.normalized_in_3d()[0])
            y = (self.radius * np.outer(np.sin(u), np.sin(v))
                 + self.center.normalized_in_3d()[1])
            z = (self.radius * np.outer(np.ones(np.size(u)), np.cos(v))
                 + self.center.normalized_in_3d()[2])
        else:
            raise ValueError("Cannot plot ball due to incompatible dimension.")

        return x, y, z



import numpy as np

from .DualQuaternion import DualQuaternion
from .PointHomogeneous import PointHomogeneous
from .RationalCurve import RationalCurve


class AffineMetric:
    """
    Class of affine metric in R12

    :references:
        M. Hofer, "Variational Motion Design in the Presence of Obstacles",
        dissertation thesis (2004), Page 7, Equation 2.4

        Schroecker, Weber, "Guaranteed collision detection with toleranced
        motions", Computer Aided Geometric Design (2014), Equation 3. DOI:
        http://dx.doi.org/10.1016/j.cagd.2014.08.001

    """
    def __init__(self, motion_curve: RationalCurve, points: list[PointHomogeneous]):
        """
        Construct the affine metric of a motion from the given points in the 3D space

        :param motion_curve: RationalCurve - rational curve representing the motion
        :param points: list[PointHomogeneous] - points in the 3D space
        """
        self.motion_curve = motion_curve
        self.points = points
        self.number_of_points = len(points)

        # By Hofer
        self.pose_distance_matrix = self.create_affine_metric()

        # By Schroecker, Weber
        self.inertia_matrix = np.sum([p[0] ** 2 * np.outer(p[1:], p[1:])
                                      for p in self.points], axis=0)
        self.inertia_eigen_vals = np.linalg.eigvals(self.inertia_matrix)
        self.total_mass = np.sum([p[0] for p in self.points])

    def __repr__(self):
        return f"{self.pose_distance_matrix}"

    def create_affine_metric(self) -> np.ndarray:
        """
        Create the affine metric of the motion

        This function computes the metric matrix for a homogeneous 3D point based on
        the formulation from M. Hofer's dissertation thesis titled "Variational Motion
        Design in the Presence of Obstacles", specifically on page 7, equation 2.4.

        :return: affine metric matrix in R12x12
        :rtype: np.ndarray

        :references:
            M. Hofer, "Variational Motion Design in the Presence of Obstacles",
            dissertation thesis (2004), Page 7, Equation 2.4

        """
        metric_matrix = np.zeros((12, 12))
        for i in range(self.number_of_points):
            metric_matrix += self.get_point_metric_matrix(self.points[i])
        return metric_matrix

    @staticmethod
    def get_point_metric_matrix(point: PointHomogeneous) -> np.ndarray:
        """
        Get the metric matrix of the given point

        :param point: PointHomogeneous - point in the 3D space

        :return: metric matrix of a single point in R12x12
        :rtype: np.ndarray

        :references:
            M. Hofer, "Variational Motion Design in the Presence of Obstacles",
            dissertation thesis (2004), Page 7, Equation 2.4

        """
        p = point.normalized_in_3d()
        i = np.eye(3)

        m00 = i
        m01 = p[0] * i
        m02 = p[1] * i
        m03 = p[2] * i

        m11 = p[0] ** 2 * i
        m12 = p[0] * p[1] * i
        m13 = p[0] * p[2] * i

        m22 = p[1] ** 2 * i
        m23 = p[1] * p[2] * i

        m33 = p[2] ** 2 * i

        metric_matrix = np.block([[m00, m01, m02, m03],
                                  [m01, m11, m12, m13],
                                  [m02, m12, m22, m23],
                                  [m03, m13, m23, m33]])

        return metric_matrix

    def get_curve_transformations(self) -> list[DualQuaternion]:
        """
        Get the transformations of the curve

        :return: transformations of the curve
        :rtype: list[DualQuaternion]
        """

        # tranformation at -1
        dq_1 = self.motion_curve.evaluate(-1)
        # transformation at infinity
        dq_inf = self.motion_curve.evaluate(0, inverted_part=True)

        return [DualQuaternion(dq_1), DualQuaternion(dq_inf)]

    def distance_via_matrix(self, a: DualQuaternion, b: DualQuaternion) -> float:
        """
        Distance between two affine displacements

        :param DualQuaternion a: displacement
        :param DualQuaternion b: displacement

        :return: distance between a and b
        :rtype: float
        """
        a12 = a.as_12d_vector()
        b12 = b.as_12d_vector()
        ab = a12 - b12
        return np.sqrt(ab @ self.pose_distance_matrix @ ab)

    def squared_distance_pr12_points(self, a: np.ndarray, b: np.ndarray) -> float:
        """
        Squared distance between two points in R12

        :param np.ndarray a: point in PR12
        :param np.ndarray b: point in PR12

        :return: squared distance between a and b
        :rtype: float
        """
        a12 = a[1:]
        b12 = b[1:]

        ab = a12 - b12
        return ab @ self.pose_distance_matrix @ ab

    def distance(self, a: DualQuaternion, b: DualQuaternion) -> float:
        """
        Distance between two affine displacements

        :param DualQuaternion a: displacement
        :param DualQuaternion b: displacement

        :return: float - distance between a and b
        :rtype: float
        """
        return np.sqrt(self.inner_product(a, b))

    def squared_distance(self, a: DualQuaternion, b: DualQuaternion) -> float:
        """
        Squared distance between two affine displacements

        :param DualQuaternion a: displacement
        :param DualQuaternion b: displacement

        :return: float - squared distance between a and b
        :rtype: float
        """
        return self.inner_product(a, b)

    def inner_product(self, a: DualQuaternion, b: DualQuaternion):
        """
        Inner product of two DualQuaternions in the affine space

        It is calculated as the sum of usual dot products of acted points, after the two
        dual quaternions act on the points that define the metric.

        :param DualQuaternion a: displacement
        :param DualQuaternion b: displacement

        :return: inner product of dq_a and dq_b
        :rtype: float
        """
        inner_product = 0
        for i in range(self.number_of_points):
            a_point = a.act(self.points[i])
            b_point = b.act(self.points[i])

            scalar = np.dot(a_point.normalized_in_3d() - b_point.normalized_in_3d(),
                            a_point.normalized_in_3d() - b_point.normalized_in_3d())
            inner_product += scalar

        return inner_product



from typing import Union
from copy import deepcopy
from warnings import warn

import sympy as sp
import numpy as np

from .DualQuaternion import DualQuaternion
from .RationalCurve import RationalCurve
from .RationalDualQuaternion import RationalDualQuaternion
from .TransfMatrix import TransfMatrix
from .PointHomogeneous import PointHomogeneous


class MotionInterpolation:
    """
    Method for interpolation of poses by rational motion curve in SE(3).

    There are two methods for interpolation of poses by rational motion curve, please
    see the following examples for more details.

    :see also: :ref:`interpolation_background`, :ref:`interpolation_examples`

    :examples:

    .. testcode::

        # 4-pose interpolation

        from rational_linkages import (DualQuaternion, Plotter, FactorizationProvider,
                                       MotionInterpolation, RationalMechanism)


        if __name__ == "__main__":
            # 4 poses
            p0 = DualQuaternion()  # identity
            p1 = DualQuaternion.as_rational([0, 0, 0, 1, 1, 0, 1, 0])
            p2 = DualQuaternion.as_rational([1, 2, 0, 0, -2, 1, 0, 0])
            p3 = DualQuaternion.as_rational([3, 0, 1, 0, 1, 0, -3, 0])

            # obtain the interpolated motion curve
            c = MotionInterpolation.interpolate([p0, p1, p2, p3])

            # factorize the motion curve
            f = FactorizationProvider().factorize_motion_curve(c)

            # create a mechanism from the factorization
            m = RationalMechanism(f)

            # create an interactive plotter object, with 1000 descrete steps
            # for the input rational curves, and arrows scaled to 0.5 length
            myplt = Plotter(interactive=True, steps=1000, arrows_length=0.5)
            myplt.plot(m, show_tool=True)

            # plot the poses
            for pose in [p0, p1, p2, p3]:
                myplt.plot(pose)

            # show the plot
            myplt.show()

    .. testcode::

        # 3-pose interpolation

        from rational_linkages import DualQuaternion, Plotter, MotionInterpolation


        if __name__ == "__main__":
            p0 = DualQuaternion([0, 17, -33, -89, 0, -6, 5, -3])
            p1 = DualQuaternion([0, 84, -21, -287, 0, -30, 3, -9])
            p2 = DualQuaternion([0, 10, 37, -84, 0, -3, -6, -3])

            c = MotionInterpolation.interpolate([p0, p1, p2])

            plt = Plotter(interactive=False, steps=500, arrows_length=0.05)
            plt.plot(c, interval='closed')

            for i, pose in enumerate([p0, p1, p2]):
                plt.plot(pose, label='p{}'.format(i+1))
    """
    def __init__(self):
        """
        Creates a new instance of the rational motion interpolation class.
        """
        pass

    @staticmethod
    def interpolate(poses: list[Union[DualQuaternion, TransfMatrix]]) -> RationalCurve:
        """
        Interpolates the given 3 poses by a rational motion curve in SE(3).

        :param list[Union[DualQuaternion, TransfMatrix]] poses: The poses to
            interpolate.

        :return: The rational motion curve.
        :rtype: RationalCurve
        """
        # check number of poses
        if not (2 <= len(poses) <= 4):
            raise ValueError('The number of poses must be 2, 3 or 4.')

        p0_array = np.asarray(poses[0].array(), dtype='float64')

        # check if the first pose is the identity matrix
        if ((isinstance(poses[0], TransfMatrix)
            and not np.allclose(p0_array, TransfMatrix().matrix))
                or (isinstance(poses[0], DualQuaternion)
                    and not np.allclose(p0_array, DualQuaternion().dq))):

            if len(poses) == 4:
                raise ValueError('The first pose must be the identity matrix '
                                 'for 4-pose interpolation')
            else:
                warn('The first pose IS NOT the identity. The interpolation '
                     'results may be unstable. They will yield non-univariate '
                     'polynomial which has to be transformed to visually '
                     'interpolate the curve.',
                     UserWarning)

        rational_poses = []

        # convert poses to rational dual quaternions
        for pose in poses:
            if isinstance(pose, TransfMatrix):
                rational_poses.append(DualQuaternion.as_rational(pose.matrix2dq()))
            elif isinstance(pose, DualQuaternion) and not pose.is_rational:
                rational_poses.append(DualQuaternion.as_rational(pose.array()))
            elif isinstance(pose, DualQuaternion) and pose.is_rational:
                rational_poses.append(pose)
            else:
                raise TypeError('The given poses must be either TransfMatrix '
                                 'or DualQuaternion.')

        # normalize the poses on Study quadric
        rational_poses = [pose.back_projection() for pose in rational_poses]

        # interpolate the rational poses
        if len(rational_poses) == 4:
            curve_eqs = MotionInterpolation.interpolate_cubic(rational_poses)
            return RationalCurve(curve_eqs)
        elif len(rational_poses) == 3:
            curve_eqs = MotionInterpolation.interpolate_quadratic(rational_poses)
            return RationalCurve(curve_eqs)
        elif len(rational_poses) == 2:
            curve_eqs = MotionInterpolation.interpolate_quadratic_2_poses(rational_poses)
            return RationalCurve(curve_eqs)

    @staticmethod
    def interpolate_quadratic(poses: list[DualQuaternion]) -> list[sp.Poly]:
        """
        Interpolates the given 3 rational poses by a quadratic curve in SE(3).

        :param list[DualQuaternion] poses: The rational poses to interpolate.

        :return: The rational motion curve.
        :rtype: list[sp.Poly]
        """
        alpha = sp.Symbol('alpha')
        omega = sp.Symbol('omega')
        t = sp.Symbol('t')

        p0 = poses[0].array()
        p1 = poses[1].array()
        p2 = poses[2].array()

        c = alpha * p2 + (p1 - alpha * p2 - omega * p0) * t + omega * p0 * t**2
        symbolic_curve = RationalDualQuaternion(c)

        # apply Stydy condition, i.e. obtain epsilon norm of the curve
        study_norm = symbolic_curve.norm()

        # simplify the norm
        study_norm = sp.simplify(study_norm[4] / (t * (t - 1)))

        # obtain the equations for alpha and omega
        eq0 = study_norm.subs(t, 0)
        eq1 = study_norm.subs(t, 1)

        # solve the equations symbolically
        sols = sp.solve([eq0, eq1], [alpha, omega], dict=True)

        # get non zero solution
        nonzero_sol = None
        for sol in sols:
            if sol[alpha] and sol[omega]:
                if (not (sol[alpha] == 0 and sol[omega] == 0)
                        and sol[alpha].is_number
                        and sol[omega].is_number):
                    nonzero_sol = sol

        if nonzero_sol is None:
            raise ValueError('Interpolation failed for the given poses.')

        al = nonzero_sol[alpha]
        om = nonzero_sol[omega]
        # obtain the interpolated curve
        c_interp = al * p2 + (p1 - al * p2 - om * p0) * t + om * p0 * t**2

        # list of polynomials
        poly = [sp.Poly(el, t) for el in c_interp]

        return poly

    @staticmethod
    def interpolate_quadratic_2_poses(poses: list[DualQuaternion]) -> list[sp.Poly]:
        """
        Interpolates the given 2 rational poses by a quadratic curve in SE(3).

        Adds the 3rd pose that is either identity or a random pose that returns
        solution.

        :param list[DualQuaternion] poses: The rational poses to interpolate.

        :return: Polynomials of rational motion curve.
        :rtype: list[sp.Poly]
        """
        try:
            return MotionInterpolation.interpolate_quadratic_2_poses_optimized(poses)
        except Exception:
            print('Failed to interpolate with a random pose optimized for shortest '
                  'path length. Trying to interpolate with other random poses...')
            return MotionInterpolation.interpolate_quadratic_2_poses_random(poses)

    @staticmethod
    def interpolate_quadratic_2_poses_random(poses: list[DualQuaternion]
                                             ) -> list[sp.Poly]:
        """
        Interpolates the given 2 rational poses by a quadratic curve in SE(3).

        Adds the 10 times 3rd pose that is random and returns the one with shortest
        path-length.

        :param list[DualQuaternion] poses: The rational poses to interpolate.

        :return: Polynomials of rational motion curve.
        :rtype: list[sp.Poly]
        """
        # Calculate the mid point between the two poses
        p0 = PointHomogeneous(poses[0].dq)
        p1 = PointHomogeneous(poses[1].dq)
        mid_p = p0.linear_interpolation(p1, 0.5)
        mid_pose = DualQuaternion(mid_p.array())

        # get mean value of mid_pose coordinates
        mean = sum(mid_pose.array()) / len(mid_pose.array())

        shortest_curve_length = float('inf')
        shortest_set = None
        best_pose = None

        for i in range(1, 10):
            additional_pose = DualQuaternion.as_rational(
                DualQuaternion.random_on_study_quadric(
                    mean * 0.3 * i).array()).back_projection()

            new_poses = deepcopy(poses)
            new_poses.append(additional_pose)

            try:
                polynomial_set = MotionInterpolation.interpolate_quadratic(
                    new_poses)
            except Exception:
                polynomial_set = None

            # If interpolation was successful, check if it's the best so far
            if polynomial_set is not None:
                new_curve_length = RationalCurve(polynomial_set).get_path_length(
                    num_of_points=500)
                if new_curve_length < shortest_curve_length:
                    shortest_set = polynomial_set
                    best_pose = additional_pose
                    shortest_curve_length = new_curve_length

        if shortest_set is not None:
            print('Chosen pose:')
            print(best_pose)

            return shortest_set

        else:  # if no solution was found
            raise ValueError('Interpolation failed for the given poses.')

    @staticmethod
    def interpolate_quadratic_2_poses_optimized(poses: list[DualQuaternion],
                                                max_iter: int = 0,
                                                ) -> list[sp.Poly]:
        """
        Interpolates the given 2 rational poses by a quadratic curve in SE(3).

        Adds the 3rd pose that is optimized for the shortest path-length.

        :param list[DualQuaternion] poses: The rational poses to interpolate
        :param int max_iter: The maximum number of iterations for the optimization,
            if 0, the optimization will run until the tolerance is reached.

        :return: Polynomials of rational motion curve.
        :rtype: list[sp.Poly]
        """
        from scipy.optimize import minimize

        # # Calculate the mid point between the two poses
        # p0 = PointHomogeneous(poses[0].dq)
        # p1 = PointHomogeneous(poses[1].dq)
        # mid_p = p0.linear_interpolation(p1, 0.5)
        # mid_pose = DualQuaternion(mid_p.array())
        # mid_pose_tr = TransfMatrix(mid_pose.dq2matrix())
        # x0 = mid_pose_tr.t
        # TODO: clean up here
        mid_pose = DualQuaternion.random_on_study_quadric()
        mid_pose_tr = TransfMatrix(mid_pose.dq2matrix())
        x0 = mid_pose_tr.t

        def objective_func(x):
            optim_pose = mid_pose_tr
            optim_pose.t = x

            new_poses = deepcopy(poses)
            new_poses.append(DualQuaternion.as_rational(
                                 optim_pose.matrix2dq()).back_projection())

            length = RationalCurve(
                MotionInterpolation.interpolate_quadratic(new_poses)).get_path_length(
                num_of_points=300
            )

            return length

        if max_iter == 0:
            res = minimize(objective_func, x0, tol=1e-3)
        else:
            res = minimize(objective_func, x0, tol=1e-3, options={'maxiter': max_iter})

        optimal_pose = mid_pose_tr
        optimal_pose.t = res.x
        optimal_pose_projected = DualQuaternion.as_rational(
            optimal_pose.matrix2dq()).back_projection()
        print('Optimal pose:')
        print(optimal_pose_projected)

        poses.append(optimal_pose_projected)

        return MotionInterpolation.interpolate_quadratic(poses)

    @staticmethod
    def interpolate_cubic(poses: list[DualQuaternion]) -> list[sp.Poly]:
        """
        Interpolates the given 4 rational poses by a cubic curve in SE(3).

        The 4 poses span a projective 3-space, which is intersected with Study quadric.
        This intersection gives another quadric containing all 4 poses, and it also
        contains cubic curves if it contains lines. The algorithm later searches
        for one of the cubic curves that interpolates the 4 poses.

        :see also: :ref:`interpolation_background`

        :param list[DualQuaternion] poses: The rational poses to interpolate.

        :return: The rational motion curve.
        :rtype: list[sp.Poly]

        :raises ValueError: If the interpolation has no solution, 'k' does not exist.
        """
        # obtain additional dual quaternions k1, k2
        try:
            k = MotionInterpolation._obtain_k_dq(poses)
        except Exception:
            raise ValueError('Interpolation has no solution.')

        # solve for t[i] - the parameter of the rational motion curve for i-th pose
        t_sols = MotionInterpolation._solve_for_t(poses, k)

        # Lagrange's interpolation part
        # lambdas for interpolation - scalar multiples of the poses
        lams = sp.symbols("lams1:5")

        parametric_points = [sp.Matrix(poses[0].array()),
                             sp.Matrix(lams[0] * poses[1].array()),
                             sp.Matrix(lams[1] * poses[2].array()),
                             sp.Matrix(lams[2] * poses[3].array())]

        # obtain the Lagrange interpolation for poses p0, p1, p2, p3
        interp = MotionInterpolation._lagrange_poly_interpolation(parametric_points)

        t = sp.symbols("t:4")
        x = sp.symbols("x")

        # to avoid calculation with infinity, substitute t[i] with 1/t[i]
        temp = [element.subs(t[0], 0) for element in interp]
        temp2 = [element.subs(x, 1 / x) for element in temp]
        temp3 = [sp.together(element * x ** 3) for element in temp2]
        temp4 = [sp.together(element.subs({t[1]: 1 / t_sols[0],
                                           t[2]: 1 / t_sols[1],
                                           t[3]: 1 / t_sols[2]}))
                 for element in temp3]

        # obtain additional parametric pose p4
        lam = sp.symbols("lam")
        poses.append(DualQuaternion([lam, 0, 0, 0, 0, 0, 0, 0]) - k[0])

        eqs_lambda = [element.subs(x, lam) - lams[-1] * poses[-1].array()[i]
                      for i, element in enumerate(temp4)]

        sols_lambda = sp.solve(eqs_lambda, lams, domain='RR')

        # obtain the family of solutions
        poly = [element.subs(sols_lambda) for element in temp4]

        # choose one solution by setting lambda, in this case lambda = 0
        poly = [element.subs(lam, 0).evalf() for element in poly]

        t = sp.Symbol("t")
        poly = [element.subs(x, t) for element in poly]

        return [sp.Poly(element, t) for element in poly]

    @staticmethod
    def _obtain_k_dq(poses: list[DualQuaternion]) -> list[DualQuaternion]:
        """
        Obtain additional dual quaternions k1, k2 for interpolation of 4 poses.

        :param list[DualQuaternion] poses: The rational poses to interpolate.

        :return: Two additional dual quaternions for interpolation.
        :rtype: list[DualQuaternion]
        """
        x = sp.symbols("x:3")

        k = DualQuaternion(poses[0].array() + x[0] * poses[1].array()
                           + x[1] * poses[2].array() + x[2] * poses[3].array())

        eqs = [k[0], k[4], k.norm().array()[4]]

        sol = sp.solve(eqs, x, domain=sp.S.Reals)

        k_as_expr = [sp.Expr(el) for el in k]

        k1 = [el.subs({x[0]: sol[0][0], x[1]: sol[0][1], x[2]: sol[0][2]})
              for el in k_as_expr]
        k2 = [el.subs({x[0]: sol[1][0], x[1]: sol[1][1], x[2]: sol[1][2]})
              for el in k_as_expr]

        k1_dq = DualQuaternion([el.args[0] for el in k1])
        k2_dq = DualQuaternion([el.args[0] for el in k2])

        return [k1_dq, k2_dq]

    @staticmethod
    def _solve_for_t(poses: list[DualQuaternion], k: list[DualQuaternion]):
        """
        Solve for t[i] - the parameter of the rational motion curve for i-th pose.

        :param list[DualQuaternion] poses: The rational poses to interpolate.
        :param list[DualQuaternion] k: The additional dual quaternions for interpolation.

        :return: The solutions for t[i].
        :rtype: list
        """
        t = sp.symbols("t:3")

        study_cond_mat = sp.Matrix([[0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0],
                                    [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 0, 1],
                                    [1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0],
                                    [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0]])

        t_dq = [DualQuaternion([t[i], 0, 0, 0, 0, 0, 0, 0]) for i in range(3)]

        eqs = [sp.Matrix((t_dq[0] - k[0]).array()).transpose() @ study_cond_mat
               @ sp.Matrix(poses[1].array()),
               sp.Matrix((t_dq[1] - k[0]).array()).transpose() @ study_cond_mat
               @ sp.Matrix(poses[2].array()),
               sp.Matrix((t_dq[2] - k[0]).array()).transpose() @ study_cond_mat
               @ sp.Matrix(poses[3].array())]

        sols_t = sp.solve(eqs, t)

        # covert to list and retrun
        return [val for i, (key, val) in enumerate(sols_t.items())]

    @staticmethod
    def _lagrange_polynomial(degree, index, x, t):
        """
        Calculate the Lagrange polynomial for interpolation.

        :param int degree: The degree of the Lagrange polynomial.
        :param int index: The index of the Lagrange polynomial.
        :param symbol x: The interpolation point (indeterminate).
        :param list[symbol] t: The interpolation nodes.

        :return: The Lagrange polynomial.
        :rtype: sp.Expr
        """
        lagrange_poly = 1
        for i in range(degree + 1):
            if i != index:
                lagrange_poly *= (x - t[i]) / (t[index] - t[i])
        return lagrange_poly

    @staticmethod
    def _lagrange_poly_interpolation(poses: list[sp.Matrix]):
        """
        Calculate the interpolation polynomial using Lagrange interpolation.

        :param list[sp.Matrix] poses: The poses to interpolate.

        :return: The interpolation polynomial.
        :rtype: sp.Matrix
        """
        # indeterminate x
        x = sp.symbols('x')

        # interpolation nodes
        t = sp.symbols("t:4")

        degree = len(poses) - 1
        result = sp.Matrix([0, 0, 0, 0, 0, 0, 0, 0])

        for i in range(degree + 1):
            result += poses[i] * MotionInterpolation._lagrange_polynomial(degree,
                                                                          i, x, t)
        return result



from typing import Union
from warnings import warn

import numpy as np


class TransfMatrix:
    """
    Transformation matrix class - following European convention

    The transformation matrix is stored as vectors n, o, a, and t, which can be changed
    independently, but be aware of the normalization of the rotation matrix.

    :ivar np.array n: normal vector (x-axis)
    :ivar np.array o: orthogonal vector (y-axis)
    :ivar np.array a: approach vector (z-axis)
    :ivar np.array t: translation vector

    :ivar np.array matrix: 4x4 transformation matrix
    """
    def __init__(self, *args):
        """
        Constructor for Trasformation Matrix class; the matrix itself is stored
        as vectors n, o, a, and t, which can be changed independently

        :param *args: empty == create identity matrix, 1 argument of matrix == SE3matrix

        :raises ValueError: if the matrix is not a proper rotation matrix
        """
        if len(args) == 0:
            mat = np.eye(4)
        else:
            mat = np.asarray(args[0])

        self.n = mat[1:4, 1]
        self.o = mat[1:4, 2]
        self.a = mat[1:4, 3]

        self.t = mat[1:4, 0]

        # test if the transformation matrix has proper rotation matrix
        if not self.is_rotation():
            raise ValueError("Matrix has not a proper rotation matrix")

    def __mul__(self, other):
        return TransfMatrix(self.matrix @ other.matrix)

    @property
    def matrix(self):
        """
        If matrix is called, return 4x4 matrix
        :return: 4x4 np array
        """
        m = np.eye(4)
        m[1:4, 1] = self.n
        m[1:4, 2] = self.o
        m[1:4, 3] = self.a

        m[1:4, 0] = self.t
        return m

    def __repr__(self):
        return np.array2string(self.matrix,
                               precision=10,
                               suppress_small=True,
                               separator=', ',
                               max_line_width=100000)

    @classmethod
    def from_rpy(cls, rpy: list[float], unit: str = 'rad') -> np.array:
        """
        Create transformation matrix from roll, pitch, yaw angles

        :param list rpy: 3-dimensional list of floats of roll, pitch, yaw angles,
            in radians or degrees in this order
        :param str unit: 'rad' or 'deg' for radians or degrees

        :return: transformation matrix
        :rtype: TransfMatrix

        :raises ValueError: if unit is not 'rad' or 'deg' or if rpy is not
            3-dimensional list
        """
        if len(rpy) != 3:
            raise ValueError("Roll, pitch, yaw angles must be 3-dimensional list of "
                             "floats")

        if unit == 'deg':
            rpy = np.deg2rad(rpy)
        elif unit != 'rad':
            raise ValueError("Unit must be 'rad' or 'deg'")

        rot_x = np.array([[1, 0, 0],
                         [0, np.cos(rpy[0]), -np.sin(rpy[0])],
                         [0, np.sin(rpy[0]), np.cos(rpy[0])]])

        rot_y = np.array([[np.cos(rpy[1]), 0, np.sin(rpy[1])],
                          [0, 1, 0],
                          [-np.sin(rpy[1]), 0, np.cos(rpy[1])]])

        rot_z = np.array([[np.cos(rpy[2]), -np.sin(rpy[2]), 0],
                          [np.sin(rpy[2]), np.cos(rpy[2]), 0],
                          [0, 0, 1]])

        m = np.eye(4)
        m[1:4, 1:4] = rot_z @ rot_y @ rot_x
        return cls(m)

    @classmethod
    def from_rpy_xyz(cls, rpy: list[float], xyz: list[float], unit: str = 'rad'):
        """
        Create transformation matrix from roll, pitch, yaw angles and translation

        :param list rpy: 3-dimensional list of floats of roll, pitch, yaw angles,
            in radians or degrees in this order
        :param list xyz: 3-dimensional list of floats of translation
        :param str unit: 'rad' or 'deg' for radians or degrees

        :return: transformation matrix
        :rtype: TransfMatrix

        :raises ValueError: if unit is not 'rad' or 'deg' or if rpy is not
            3-dimensional list
        """
        if len(rpy) != 3 or len(xyz) != 3:
            raise ValueError("Roll, pitch, yaw angles or XYZ valuse must "
                             "be 3-dimensional list of floats")

        mat_applied_rotation = cls.from_rpy(rpy, unit)

        # update translation
        mat_applied_rotation.t = xyz

        return cls(mat_applied_rotation.matrix)

    @classmethod
    def from_vectors(cls,
                     normal_x: list[Union[float, np.ndarray]],
                     approach_z: list[Union[float, np.ndarray]],
                     origin: list[Union[float, np.ndarray]] = [0, 0, 0]):
        """
        Create transf. matrix from normal and approach vectors, translation is optional

        :param list normal_x: 3-dimensional list of floats of normal (x-axis) vector
        :param list approach_z: 3-dimensional list of floats of approach (z-axis) vector
        :param list origin: 3-dimensional list of floats of translation vector

        :return: transformation matrix
        :rtype: TransfMatrix

        :raises ValueError: if normal_x, approach_z or origin is not 3-dimensional list
        :warns: if normal_x or approach_z is not normalized
        """
        normal_x = np.asarray(normal_x)
        approach_z = np.asarray(approach_z)
        origin = np.asarray(origin)

        # check if vectors are of dimension 3
        if normal_x.shape != (3,):
            raise ValueError("Normal vector must be 3-dimensional list of floats")
        if approach_z.shape != (3,):
            raise ValueError("Approach vector must be 3-dimensional list of floats")
        if origin.shape != (3,):
            raise ValueError("Origin vector must be 3-dimensional list of floats")

        # check if approach vector is normalized
        if not np.isclose(np.linalg.norm(approach_z), 1):
            warn("Approach vector is not normalized, normalizing...")
            approach_z = approach_z / np.linalg.norm(approach_z)

        # create orthogonal
        orthogonal_y = np.cross(approach_z, normal_x)
        # recreate normal (only approach vector can be kept so det(R) == 1)
        normal_x = np.cross(orthogonal_y, approach_z)

        # normalize orthogonal and normal vectors
        orthogonal_y = orthogonal_y / np.linalg.norm(orthogonal_y)
        normal_x = normal_x / np.linalg.norm(normal_x)

        mat = np.eye(4)
        mat[1:4, 0] = origin
        mat[1:4, 1] = normal_x
        mat[1:4, 2] = orthogonal_y
        mat[1:4, 3] = approach_z

        return cls(mat)

    @classmethod
    def from_dh_parameters(cls, theta: float, d: float, a: float, alpha: float,
                           unit: str = 'rad'):
        """
        Create transformation matrix from Denavit-Hartenberg parameters

        It follows the standard DH convention. The transformation matrix is created as:
        rotation around z-axis by theta, translation along z-axis by d, translation
        along x-axis by a, rotation around x-axis by alpha.

        :param float theta: rotation around z-axis
        :param float d: translation along z-axis
        :param float a: translation along x-axis
        :param float alpha: rotation around x-axis
        :param str unit: 'rad' or 'deg' for radians or degrees

        :return: transformation matrix
        :rtype: TransfMatrix

        :raises ValueError: if unit is not 'rad' or 'deg'
        """
        if unit == 'deg':
            theta = np.deg2rad(theta)
            alpha = np.deg2rad(alpha)
        elif unit != 'rad':
            raise ValueError("Unit must be 'rad' or 'deg'")

        mat = np.eye(4)
        mat[1:4, 0] = [a * np.cos(theta), a * np.sin(theta), d]
        mat[1, 1:4] = [np.cos(theta),
                       -np.sin(theta) * np.cos(alpha),
                       np.sin(theta) * np.sin(alpha)]
        mat[2, 1:4] = [np.sin(theta),
                       np.cos(theta) * np.cos(alpha),
                       -np.cos(theta) * np.sin(alpha)]
        mat[3, 1:4] = [0, np.sin(alpha), np.cos(alpha)]

        return cls(mat)

    def array(self) -> np.array:
        """
        Return transformation matrix as 4x4 numpy array

        :return: 4x4 numpy array
        :rtype: np.array
        """
        return self.matrix

    def is_rotation(self):
        """
        Check if matrix is rotation matrix with determinant equal to 1

        :return: True if matrix is rotation
        """
        if np.isclose(np.linalg.det(self.rot_matrix()), 1):
            return True
        else:
            warn("Matrix has NOT determinant equal to 1")
            return False

    def matrix2dq(self) -> np.array:
        """
        Convert SE(3) matrix representation to dual quaternion array

        :return: return 8-dimensional array of dual quaternion
        :rtype: np.array
        """
        conditions = [
            (1 + self.n[0] + self.o[1] + self.a[2],
             self.o[2] - self.a[1],
             self.a[0] - self.n[2],
             self.n[1] - self.o[0]),
            (self.o[2] - self.a[1],
             1 + self.n[0] - self.o[1] - self.a[2],
             self.o[0] + self.n[1],
             self.n[2] + self.a[0]),
            (self.a[0] - self.n[2],
             self.o[0] + self.n[1],
             1 - self.n[0] + self.o[1] - self.a[2],
             self.a[1] + self.o[2]),
            (self.n[1] - self.o[0],
             self.n[2] + self.a[0],
             self.a[1] + self.o[2],
             1 - self.n[0] - self.o[1] + self.a[2])
        ]

        p = np.zeros(4)
        for condition in conditions:
            p[0], p[1], p[2], p[3] = condition
            if p[0] != 0 or sum(p) != 0:
                break

        d = np.zeros(4)
        d[0] = (self.t[0] * p[1] + self.t[1] * p[2] + self.t[2] * p[3]) / 2
        d[1] = (-self.t[0] * p[0] + self.t[2] * p[2] - self.t[1] * p[3]) / 2
        d[2] = (-self.t[1] * p[0] - self.t[2] * p[1] + self.t[0] * p[3]) / 2
        d[3] = (-self.t[2] * p[0] + self.t[1] * p[1] - self.t[0] * p[2]) / 2

        # normalization
        if p[0] != 0:
            d = d / p[0]
            p = p / p[0]
        else:
            norm = np.linalg.norm(p)
            d = d / norm
            p = p / norm

        return np.concatenate((p, d))

    def rot_matrix(self):
        r = self.matrix[1:4, 1:4]
        return r

    def rpy(self):
        """
        Return roll, pitch, yaw angles of the rotation matrix
        
        :return: 3-dimensional numpy array of roll, pitch, yaw angles
        """
        rpy = np.zeros((3,))
        R = self.rot_matrix()
        if abs(abs(R[2, 0]) - 1) < 10 * np.finfo(np.float64).eps:  # when |R31| == 1
            # singularity
            rpy[0] = 0  # roll is zero
            if R[2, 0] < 0:
                rpy[2] = -np.arctan2(R[0, 1], R[0, 2])  # R-Y
            else:
                rpy[2] = np.arctan2(-R[0, 1], -R[0, 2])  # R+Y
            rpy[1] = -np.arcsin(np.clip(R[2, 0], -1.0, 1.0))
        else:
            rpy[0] = np.arctan2(R[2, 1], R[2, 2])  # R
            rpy[2] = np.arctan2(R[1, 0], R[0, 0])  # Y

            k = np.argmax(np.abs([R[0, 0], R[1, 0], R[2, 1], R[2, 2]]))
            if k == 0:
                rpy[1] = -np.arctan(R[2, 0] * np.cos(rpy[2]) / R[0, 0])
            elif k == 1:
                rpy[1] = -np.arctan(R[2, 0] * np.sin(rpy[2]) / R[1, 0])
            elif k == 2:
                rpy[1] = -np.arctan(R[2, 0] * np.sin(rpy[0]) / R[2, 1])
            elif k == 3:
                rpy[1] = -np.arctan(R[2, 0] * np.cos(rpy[0]) / R[2, 2])

        return rpy

    def dh_to_other_frame(self, other: 'TransfMatrix') -> list[float]:
        """
        Return Denavit-Hartenberg parameters from one frame to another

        :param TransfMatrix other: transformation matrix of the other frame

        :return: DH parameters [theta, d, a, alpha]
        :rtype: list[float]

        :warns: if the two frames do not fulfill the DH convention
        """
        from .NormalizedLine import NormalizedLine

        # theta
        th = np.arctan2(np.linalg.norm(np.cross(self.n, other.n)),
                        np.dot(self.n, other.n))
        # using right-hand rule, if a dot product of Xi axis is in negative
        theta = -th if np.dot(self.o, other.n) < 0 else th

        # d, using normal vectors (x-axis) of the frames
        line0 = NormalizedLine.from_direction_and_point(self.n, self.t)
        line1 = NormalizedLine.from_direction_and_point(other.n, other.t)

        _, dist, __ = line0.common_perpendicular_to_other_line(line1)

        # using right-hand rule, if the dot product of vectors between two links and
        # z-axis is less than 0 it has to be different direction
        d = -dist if np.dot(other.t - self.t, self.a) < 0 else dist

        # a, using approach vectors (z-axis) of the frames
        line0 = NormalizedLine.from_direction_and_point(self.a, self.t)
        line1 = NormalizedLine.from_direction_and_point(other.a, other.t)

        _, dist, __ = line0.common_perpendicular_to_other_line(line1)

        # using right-hand rule, if the dot product of vectors between two links and
        # x-axis is less than 0 it has to be different direction
        a = -dist if np.dot(other.t - self.t, other.n) < 0 else dist

        # alpha
        al = np.arctan2(np.linalg.norm(np.cross(self.a, other.a)),
                        np.dot(self.a, other.a))

        # using right-hand rule, if a dot product of Zi-1 axis is in negative direction
        # of Yi axis, the angle has to be negative
        alpha = -al if np.dot(other.o, self.a) < 0 else al

        # check if the two frames fulfill the DH convention
        test_frame = TransfMatrix.from_dh_parameters(theta, d, a, alpha)
        if not np.allclose(self.matrix @ test_frame.matrix, other.matrix):
            warn("Frames do not fulfill the DH convention")

        return [theta, d, a, alpha]

    def inv(self):
        """
        Return inverse transformation matrix

        :return: inverse transformation matrix
        :rtype: TransfMatrix
        """
        inv_rotation = np.transpose(self.rot_matrix())
        inv_translation = -inv_rotation @ self.t

        m = np.eye(4)
        m[1:4, 1:4] = inv_rotation
        m[1:4, 0] = inv_translation
        return TransfMatrix(m)

    def get_plot_data(self):
        """
        Return three quiver coordinates for plotting

        :return: 6-dimensional numpy array of point and vector direction
        """
        x_vec = np.concatenate((self.t, self.n))
        y_vec = np.concatenate((self.t, self.o))
        z_vec = np.concatenate((self.t, self.a))

        return x_vec, y_vec, z_vec



from matplotlib.widgets import Button


class ButtonID(Button):
    def __init__(
        self, ax, label, image=None, color="0.85", hovercolor="0.95", button_id=1
    ):
        """
        Button class derivative to support an ID assignment to a button. ID then
        corresponds a joint order.

        :param ax:
        :param label:
        :param image:
        :param color:
        :param hovercolor:
        :param button_id:
        """
        Button.__init__(self, ax, label, image, color, hovercolor)
        self.func = None
        self.button_id = button_id

    def on_clicked(self, func):
        """
        Method override for "on_clicked" function. func is saved and handled later
        in "_clicked" method together
        with the ID.

        :param func: original func
        :return:
        """
        Button.on_clicked(self, self._clicked)
        self.func = func

    def _clicked(self, event):
        """
        Method that adds ID to the func values.

        :param event:
        :return:
        """
        self.func(event, self.button_id)



from typing import Optional, Sequence

import numpy as np


class Quaternion:
    """
    Quaternion class representing a 4-dimensional quaternion.

    :ivar np.ndarray q: 4-vector of quaternion parameters

    :examples:

    .. testcode::

        # General usage

        from rational_linkages import Quaternion

        identity_quaternion = Quaternion()
        quaternion_from_list = Quaternion([0.5, 2, 1, 5])
    """

    def __init__(self, vec4: Optional[Sequence[float]] = None):
        """
        Quaternion class

        :param Optional[Sequence[float]] vec4: 4-vector list of quaternion parameters
        """
        if vec4 is not None:
            if len(vec4) != 4:
                raise ValueError("Quaternion: vec4 has to be 4-vector")
            self.q = np.asarray(vec4)
        else:
            self.q = np.array([1, 0, 0, 0])

        self.real = self.q[0]
        self.imag = self.q[1:]

    def __getitem__(self, idx):
        """
        Get an element of Quaternion

        :param idx: index of the Quaternion element to call 0..3
        :return: float
        """
        element = self.q
        element = element[idx]  # or, p.dob = p.dob.__getitem__(indx)
        return element

    def __repr__(self):
        """
        Printing method override

        :return: Quaterion in readable form
        :rtype: str
        """
        q = np.array2string(self.array(),
                            precision=10,
                            suppress_small=True,
                            separator=', ',
                            max_line_width=100000)
        return f"{self.__class__.__qualname__}({q})"

    def __add__(self, other):
        """
        Quaternion addition

        :param other: Quaternion
        :return: Quaternion
        """
        return Quaternion(self.q + other.q)

    def __sub__(self, other):
        """
        Quaternion subtraction

        :param other: Quaternion
        :return: Quaternion
        """
        return Quaternion(self.q - other.q)

    def __mul__(self, other):
        """
        Quaternion Multiplication

        :param other: Quaternion
        :return: Quaternion
        """
        if isinstance(other, (int, float)):
            return Quaternion(self.q * other)
        else:
            w, x, y, z = self.q
            ow, ox, oy, oz = other.q
            return Quaternion(np.array([w * ow - x * ox - y * oy - z * oz,
                                        w * ox + x * ow + y * oz - z * oy,
                                        w * oy - x * oz + y * ow + z * ox,
                                        w * oz + x * oy - y * ox + z * ow]))

    def __rmul__(self, other):
        """Handle when the quaternion is on the right side of the multiplication"""
        return self.__mul__(other)

    def __truediv__(self, other):
        """
        Quaternion division

        :param Quaternion, int, float other: Quaternion

        :return: Quaternion
        :rtype: Quaternion
        """
        if isinstance(other, (int, float)):
            return Quaternion(self.q / other)
        else:
            return self * other.inv()

    def __neg__(self):
        """
        Quaternion negation

        :return: Quaternion
        :rtype: Quaternion
        """
        return Quaternion(-1 * self.q)

    def __eq__(self, other):
        """
        Compare two Quaternions if they are equal

        :param other: Quaternion
        :return: bool
        """
        return np.array_equal(self.array(), other.array())

    def array(self):
        """
        Quaternion to numpy array

        :return: numpy array of quaternion 4-vector parameters
        :rtype: np.ndarray
        """
        return np.array([self.q[0], self.q[1], self.q[2], self.q[3]])

    def conjugate(self):
        """
        Quaternion conjugate

        :return: Quaternion
        :rtype: Quaternion
        """
        q0 = self.q[0]
        q1 = -1 * self.q[1]
        q2 = -1 * self.q[2]
        q3 = -1 * self.q[3]

        return Quaternion(np.array([q0, q1, q2, q3]))

    def norm(self):
        """
        Quaternion norm, aslo called the Quadrance

        :return: Quaternion
        :rtype: float
        """
        q0 = self.q[0]
        q1 = self.q[1]
        q2 = self.q[2]
        q3 = self.q[3]

        return q0**2 + q1**2 + q2**2 + q3**2

    def length(self):
        """
        Quaternion length

        :return: Quaternion
        :rtype: float
        """
        return np.sqrt(self.norm())

    def inv(self):
        """
        Quaternion inverse

        :return: Quaternion
        :rtype: Quaternion
        """
        return Quaternion(self.conjugate().array() / self.norm())



import numpy as np
from scipy.optimize import minimize

from .DualQuaternion import DualQuaternion


class MotionApproximation:
    """
    MotionApproximation class
    """

    def __init__(self, poses: list[DualQuaternion]):
        """
        Initializes a MotionApproximation object

        :param poses: list[DualQuaternion] - list of poses
        """
        self.poses = poses
        self.num_poses = len(poses)
        self.num_params = int(8 * 3 + 2)  # for degree 2 curve

        self.curve = self.get_curve()

    def get_curve(self):
        """
        Get the curve of the motion approximation

        :return: MotionApproximationCurve
        """

        initial_guess = np.zeros(self.num_params)
        initial_guess[2] = 1.0

        def objective_function(params):
            """
            Objective function to minimize the sum of squared distances between
            the poses and the curve
            """
            #t, s0, s1, s2, i0, i1, i2, j0, j1, j2, k0, k1, k2, es0, es1, es2, ei0, ei1, ei2, ej0, ej1, ej2, ek0, ek1, ek2 = params
            return np.sum([val ** 2 for val in params[2::]])

        def constraint(params):
            t0, t1, s0, s1, s2, i0, i1, i2, j0, j1, j2, k0, k1, k2, es0, es1, es2, ei0, ei1, ei2, ej0, ej1, ej2, ek0, ek1, ek2 = params
            t = [t0, t1]

            #c0 = ((s0 + s1 * t + s2 * t ** 2) * (es0 + es1 * t + es2 * t ** 2)
            #      + (i0 + i1 * t + i2 * t ** 2) * (ei0 + ei1 * t + ei2 * t ** 2)
            #      + (j0 + j1 * t + j2 * t ** 2) * (ej0 + ej1 * t + ej2 * t ** 2)
            #      + (k0 + k1 * t + k2 * t ** 2) * (ek0 + ek1 * t + ek2 * t ** 2))

            c = []

            for i in range(self.num_poses):
                c.append(s0 + s1 * t[i] + s2 * t[i] ** 2 - self.poses[i].dq[0])
                c.append(i0 + i1 * t[i] + i2 * t[i] ** 2 - self.poses[i].dq[1])
                c.append(j0 + j1 * t[i] + j2 * t[i] ** 2 - self.poses[i].dq[2])
                c.append(k0 + k1 * t[i] + k2 * t[i] ** 2 - self.poses[i].dq[3])
                c.append(es0 + es1 * t[i] + es2 * t[i] ** 2 - self.poses[i].dq[4])
                c.append(ei0 + ei1 * t[i] + ei2 * t[i] ** 2 - self.poses[i].dq[5])
                c.append(ej0 + ej1 * t[i] + ej2 * t[i] ** 2 - self.poses[i].dq[6])
                c.append(ek0 + ek1 * t[i] + ek2 * t[i] ** 2 - self.poses[i].dq[7])

            return c

        constraints = {'type': 'eq', 'fun': constraint}

        result = minimize(objective_function, initial_guess, constraints=constraints, method='SLSQP')

        optimized_params = result.x
        print("Optimized Parameters:", optimized_params)
        print(result)

        return result






from typing import Union
from warnings import warn

import biquaternion_py
import numpy as np
import sympy as sp

from .DualQuaternion import DualQuaternion
from .MotionFactorization import MotionFactorization
from .RationalCurve import RationalCurve


class FactorizationProvider:
    """
    This class provides the factorizations for the given curve or motion factorization.

    It connetion to the project BiQuaternions_py made by Daren Thimm, University of
    Innbruck, Austria. Git repository: `BiQuaternions_py`_.

    .. _BiQuaternions_py: https://git.uibk.ac.at/geometrie-vermessung/biquaternion_py
    """
    def __init__(self, use_rationals: bool = False):
        """
        Creates a new instance of the FactorizationProvider class.

        :param bool use_rationals: If True, the factorization will be performed
            with rational numbers in QQ instead of floating point numbers in RR.

        :ivar str domain: The domain of the factorization, either 'QQ' or 'RR'.
        """
        self.domain = 'QQ' if use_rationals else 'RR'

    def factorize_motion_curve(self,
                               curve: Union[RationalCurve,
                               biquaternion_py.polynomials.Poly]) -> (
            list)[MotionFactorization]:
        """
        Factorizes the given curve into a multiple motion factorizations.

        :param Union[RationalCurve, biquaternion_py.polynomials.Poly] curve: The curve
            to factorize, either as a RationalCurve or as a polynomial.

        :return: The factorizations of the curve.
        :rtype: list[MotionFactorization]

        :warning: If the given curve has not only rational numbers as input.
        """
        t = sp.Symbol("t")

        if isinstance(curve, RationalCurve):
            bi_quat = biquaternion_py.BiQuaternion(curve.extract_expressions())
            bi_poly = biquaternion_py.polynomials.Poly(bi_quat, t)
        else:
            bi_poly = curve

        # check if the given curve has rational numbers as input
        poly_coeffs = bi_poly.all_coeffs()
        for i in range(len(poly_coeffs)):
            for j in range(len(poly_coeffs[i].args)):
                if not isinstance(poly_coeffs[i].args[j], sp.Rational):
                    warn('The given curve has not only rational numbers as input. The factorization will be performed with floating point numbers, but may be instable.')
                    break

        factorizations = self.factorize_polynomial(bi_poly)

        factors1 = [self.factor2rotation_axis(factor) for factor in factorizations[0]]
        factors2 = [self.factor2rotation_axis(factor) for factor in factorizations[1]]

        return [MotionFactorization(factors1), MotionFactorization(factors2)]

    def factorize_for_motion_factorization(self, factorization: MotionFactorization) \
            -> list[MotionFactorization]:
        """
        Analyzes the given motion factorization and provides other motion
        factorizations, if possible.

        :param MotionFactorization factorization: The motion factorization to
            factorize for.

        :return: The factorizations of the motion factorization.
        :rtype: list[MotionFactorization]

        :warning: If the given motion factorization has not only dual
            quaternions with rational numbers elements as input.
        """
        # check if the given factorization has input DualQuaternions as rational numbers
        for i in range(factorization.number_of_factors):
            if not factorization.dq_axes[i].is_rational:
                warn('The given motion factorization has not only rational numbers as input. The factorization will be performed with floating point numbers, but may be instable.')

        t = sp.Symbol("t")

        bi_poly = t - biquaternion_py.BiQuaternion(factorization.dq_axes[0].array())
        for i in range(1, factorization.number_of_factors):
            bi_poly = bi_poly * (t - biquaternion_py.BiQuaternion(factorization.dq_axes[i].array()))

        bi_poly = biquaternion_py.polynomials.Poly(bi_poly, t)

        return self.factorize_motion_curve(bi_poly)

    def factorize_polynomial(self,
                             poly: biquaternion_py.polynomials.Poly) -> (
            list)[biquaternion_py.polynomials.Poly]:
        """
        Factorizes the given polynomial into irreducible factors.

        :param biquaternion_py.polynomials.Poly poly: The polynomial to factorize.

        :return: The irreducible factors of the polynomial.
        :rtype: list[biquaternion_py.polynomials.Poly]

        :raises: If the factorization failed.
        """
        # Calculate the norm polynomial. To avoid numerical problems, extract
        # the scalar part, since the norm should be purely real
        norm_poly = poly.norm()
        norm_poly = biquaternion_py.polynomials.Poly(norm_poly.poly.scal,
                                                     *norm_poly.indets)

        # Calculate the irreducible factors, that determine the different factorizations
        _, factors = biquaternion_py.irreducible_factors(norm_poly, domain=self.domain)

        # The different permutations of the irreducible factors then generate
        # the different factorizations of the motion.

        if len(factors) <= 1:
            raise ValueError('The factorization failed for the given input.')

        factorization1 = biquaternion_py.factorize_from_list(poly, factors)
        factorization2 = biquaternion_py.factorize_from_list(poly, factors[::-1])

        return [factorization1, factorization2]

    def factor2rotation_axis(self,
                             factor: biquaternion_py.polynomials.Poly) -> (
            DualQuaternion):
        """
        Converts the given factor to a dual quaternion representing the rotation axis
        of a linkage, excluding the parameter.

        :param biquaternion_py.polynomials.Poly factor: The factor to convert.

        :return: The rotation axis of the factor.
        :rtype: DualQuaternion
        """
        t = sp.Symbol("t")
        t_dq = DualQuaternion([t, 0, 0, 0, 0, 0, 0, 0])

        factor_dq = DualQuaternion(factor.poly.coeffs)

        # subtract the parameter from the factor
        axis_h = t_dq - factor_dq

        if self.domain == 'QQ':
            return DualQuaternion(axis_h.array())
        else:
            axis_h = np.asarray(axis_h.array(), dtype='float64')
            return DualQuaternion(axis_h)






import numpy as np

from .TransfMatrix import TransfMatrix

# Forward declarations for class names
DualQuaternion = "DualQuaternion"


class PointHomogeneous:
    """
    Points in projective space with homogeneous coordinates.

    Homogeneous coordinates are used to represent points, including points at infinity.
    The first row of the point array (index 0) stores the homogeneous coordinates.

    :ivar coordinates: Array of floats representing the homogeneous coordinates of the
        point.
    :ivar is_at_infinity: Indicates whether the point is at infinity (coordinates[0] is
        close to 0).

    :examples:

    .. testcode::

        # Create points in projective space

        from rational_linkages import PointHomogeneous


        origin_point_3D = PointHomogeneous()
        origin_point_2D = PointHomogeneous.at_origin_in_2d()
        custom_point = PointHomogeneous([2.0, 3.0, 4.0, 1.0])
    """

    def __init__(self, point=None):
        """
        Class to store points in PR3 or PR2

        Homogeneous coordinates are stored in the first row of the point array (index 0)
        :param point: array or list of floats
        """
        from sympy import Expr

        self.is_real = True

        if point is None:  # point in the origin in PR3
            self.coordinates = np.array([1, 0, 0, 0])
        elif any(isinstance(element, Expr) for element in point):
            try:
                self.coordinates = np.asarray(point, dtype='float64')
            except Exception:
                self.coordinates = point
                self.is_real = False
        else:
            self.coordinates = np.asarray(point, dtype='float64')

        if self.is_real and np.allclose(np.array([self.coordinates[0]]), np.array([0.0])):  # point at infinity
            self.is_at_infinity = True
            self.coordinates_normalized = None
        elif self.is_real:
            self.is_at_infinity = False
            self.coordinates_normalized = self.normalize()
        else:
            self.is_at_infinity = False
            self.coordinates_normalized = None

        #if len(self.coordinates_normalized) == 4:  # point in PR3
        #    self.as_dq_array = self.point2dq_array()

        self.orbit = None

    @classmethod
    def at_origin_in_2d(cls):
        """
        Create homogeneous point at origin in 2D

        :return: PointHomogeneous
        """
        point = np.zeros(3)
        point[0] = 1
        return cls(point)

    @classmethod
    def from_3d_point(cls, point: np.ndarray):
        """
        Create homogeneous point from 3D point

        :param point: 3D point
        :return: PointHomogeneous
        """
        point = np.asarray(point)
        if len(point) != 3:
            raise ValueError("PointHomogeneous: point has to be 3D")
        point = np.insert(point, 0, 1)
        return cls(point)

    @classmethod
    def from_dual_quaternion(cls, dq: "DualQuaternion"):
        """
        Create homogeneous point from dual quaternion

        :param dq: DualQuaternion
        :return: PointHomogeneous
        """

        p0 = dq[0]
        p1 = dq[5]
        p2 = dq[6]
        p3 = dq[7]
        return cls([p0, p1, p2, p3])

    def __getitem__(self, idx):
        """
        Get specified coordinate from point
        :param idx: coordinate index
        :return: float
        """
        return self.coordinates[idx]

    def __repr__(self):
        """
        Print point
        :return:
        """
        p = np.array2string(self.array(),
                            precision=10,
                            suppress_small=True,
                            separator=', ',
                            max_line_width=100000)
        return f"{p}"

    def __add__(self, other):
        """
        Add two points
        :param other: Point
        :return: array of floats
        """
        return PointHomogeneous(self.coordinates + other.coordinates)

    def __sub__(self, other):
        """
        Subtract two points
        :param other: Point
        :return: array of floats
        """
        return PointHomogeneous(self.coordinates - other.coordinates)

    def array(self) -> np.ndarray:
        """
        Return point as numpy array
        :return:
        """
        return self.coordinates

    def normalize(self) -> np.ndarray:
        """
        Normalize the point
        :return: 4x1 array
        """
        if self.is_at_infinity:
            return self.coordinates / np.linalg.norm(self.coordinates)
        else:
            return self.coordinates / self.coordinates[0]

    def normalized_in_3d(self) -> np.ndarray:
        """
        Normalize the point and return its 3D coordinates
        :return: 3x1 array
        """
        return self.coordinates_normalized[1:]

    def point2matrix(self) -> np.ndarray:
        """
        Convert point to homogeneous SE3 matrix with identity in rotation part

        This methods follows the European convention for SE3 matrices, i.e. the first
        column of the matrix is translation and the rotation part is represented by
        the remaining 3 columns.

        :return: 4x4 array
        :rtype: np.ndarray
        """
        mat = np.eye(4)
        if len(self.coordinates_normalized) == 3:  # point in PR2
            mat[1:3, 0] = self.coordinates_normalized[1:3]
        elif len(self.coordinates_normalized) == 4:
            mat[1:4, 0] = self.coordinates_normalized[1:4]

        # affine displacement in R12
        elif len(self.coordinates_normalized) == 12:
            mat[1:4, 0] = self.coordinates_normalized[0:3]
            mat[1:4, 1] = self.coordinates_normalized[3:6]
            mat[1:4, 2] = self.coordinates_normalized[6:9]
            mat[1:4, 3] = self.coordinates_normalized[9:12]

        # affine displacement in PR12
        elif len(self.coordinates_normalized) == 13:
            mat[1:4, 0] = self.coordinates_normalized[1:4]
            mat[1:4, 1] = self.coordinates_normalized[4:7]
            mat[1:4, 2] = self.coordinates_normalized[7:10]
            mat[1:4, 3] = self.coordinates_normalized[10:13]
        else:
            raise ValueError("PointHomogeneous: point has to be in PR2 or PR3")

        return mat

    def point2dq_array(self) -> np.ndarray:
        """
        Embed point to dual quaternion space

        :return: np.array of shape (8,)
        """
        return np.array(
            [
                self.coordinates[0],
                0,
                0,
                0,
                0,
                self.coordinates[1],
                self.coordinates[2],
                self.coordinates[3],
            ]
        )

    def point2affine12d(self, map_alpha: TransfMatrix) -> np.array:
        """
        Map point to 12D affine space

        :param TransfMatrix map_alpha: SE3matrix object that maps point to 12D affine
            space

        :return: 12D affine point
        :rtype: np.array
        """

        x = self.coordinates
        point0 = x[0] * map_alpha.t

        point1 = x[1] * map_alpha.n
        point2 = x[2] * map_alpha.o
        point3 = x[3] * map_alpha.a

        return np.concatenate((point0, point1, point2, point3))

    def linear_interpolation(self, other, t: float = 0.5) -> "PointHomogeneous":
        """
        Linear interpolation between two points

        :param other: PointHomogeneous
        :param t: parameter of interpolation in range [0, 1]
        :return: PointHomogeneous
        """
        return PointHomogeneous(self.coordinates * (1 - t) + other.coordinates * t)

    def get_plot_data(self) -> np.ndarray:
        """
        Get data for plotting in 3D space

        :return: np.ndarray of shape (3, 1)
        """
        return self.normalized_in_3d()

    def evaluate(self, t_param: float) -> 'PointHomogeneous':
        """
        Evaluate the point at the given parameter

        :param float t_param: parameter

        :return: evaluated point with float elements
        :rtype: PointHomogeneous
        """
        from sympy import Expr, Symbol, Number

        t = Symbol("t")

        point_expr = [Expr(coord) if not isinstance(coord, Number) else coord
                      for coord in self.coordinates]
        point = [coord.subs(t, t_param).evalf().args[0]
                 if not isinstance(coord, Number) else coord
                 for coord in point_expr]
        return PointHomogeneous(np.asarray(point, dtype="float64"))

    def get_point_orbit(self,
                        acting_center: "PointHomogeneous",
                        acting_radius: float,
                        metric: "AffineMetric",
                        ) -> tuple[np.ndarray, float]:
        """
        Get point orbit



        :param PointHomogeneous acting_center: center of the acting ball
        :param float acting_radius: radius of the orbit ball
        :param AffineMetric metric: metric of the curve

        :return: point center and radius squared
        :rtype: tuple[np.ndarray, float]
        """
        point_center = acting_center.point2matrix() @ self.coordinates_normalized

        coords_3d = self.normalized_in_3d()
        radius_squared = acting_radius ** 2 * (1/metric.total_mass + np.sum([(coord ** 2 / metric.inertia_eigen_vals[i]) for i, coord in enumerate(coords_3d)]))

        radius = np.sqrt(radius_squared)
        self.set_point_orbit(point_center, radius)
        return point_center, radius

    def set_point_orbit(self, orbit_center: "PointHomogeneous", orbit_radius: float):
        """
        Set the orbit of the point
        """
        self.orbit = PointOrbit(orbit_center, orbit_radius)


class PointOrbit:
    def __init__(self, point_center, radius):
        """

        """
        if not isinstance(point_center, PointHomogeneous):
            self.center = PointHomogeneous(point_center)
        else:
            self.center = point_center

        self.radius = radius

    def get_plot_data(self) -> tuple:
        """
        Get data for plotting in 3D space

        :return: surface coordinates
        :rtype: tuple
        """
        if len(self.center.coordinates) == 4:
            # Create the 3D sphere representing the circle
            u = np.linspace(0, 2 * np.pi, 10)
            v = np.linspace(0, np.pi, 10)

            x = (self.radius * np.outer(np.cos(u), np.sin(v))
                 + self.center.normalized_in_3d()[0])
            y = (self.radius * np.outer(np.sin(u), np.sin(v))
                 + self.center.normalized_in_3d()[1])
            z = (self.radius * np.outer(np.ones(np.size(u)), np.cos(v))
                 + self.center.normalized_in_3d()[2])
        else:
            raise ValueError("Cannot plot ball due to incompatible dimension.")

        return x, y, z



from .RationalMechanism import RationalMechanism
from .NormalizedLine import NormalizedLine

import numpy as np
from itertools import product


class CollisionFreeOptimization:
    """
    Class for the optimization of the mechanism for full-cycle collision-free design.
    """
    def __init__(self, mechanism: RationalMechanism):
        """
        Initialize the combinatorial search algorithm.

        :param RationalMechanism mechanism: The mechanism to optimize.
        """
        self.mechanism = mechanism

    def smallest_polyline(self) -> tuple:
        """
        Get points on mechanism axes that form the smallest polyline.

        This method calculates the smallest polyline that can be formed by points on
        the mechanism axes. It uses scipy's minimize function to find the points on
        the axes that minimize the total distance of the polyline.

        :return: points on the mechanism axes that form the smallest polyline,
            parameters of the points, result of the optimization
        :rtype: tuple
        """
        from scipy.optimize import minimize

        # get the axes represented as normalized lines
        dq_lines = (self.mechanism.factorizations[0].dq_axes
                    + self.mechanism.factorizations[1].dq_axes[::-1])
        lines = [NormalizedLine.from_dual_quaternion(dq_line) for dq_line in dq_lines]

        def objective_function(x):
            p = [line.point_on_line(x[i]) for i, line in enumerate(lines)]

            total_distance = sum(
                np.linalg.norm(p[i] - p[i + 1])
                for i in range(self.mechanism.num_joints - 1))
            # Add distance between last and first point
            total_distance += np.linalg.norm(p[-1] - p[0])
            return total_distance

        x_init = np.zeros(self.mechanism.num_joints)
        result = minimize(objective_function, x_init)

        # double the parameters for the two joint connection points
        points_params = result.x
        points_params = [np.array([param, param]) for param in points_params]

        points = [line.point_on_line(float(points_params[i][0]))
                  for i, line in enumerate(lines)]

        return points, points_params, result

    def optimize(self,
                 method: str,
                 step_length: float,
                 min_joint_segment_length: float,
                 max_iters: int,
                 **kwargs):
        """
        Optimize the mechanism for collision-free operation.

        :param method: optimization method
        :param step_length: length of the step, i.e. the shift distance value, see
            :ref:`combinatorial_search` for more detail
        :param min_joint_segment_length: minimum length of the joint segment
        :param max_iters: maximum number of iterations
        :param kwargs: additional keyword arguments
        """
        # initial estimation - the smallest polyline
        points, points_params, result = self.smallest_polyline()

        # update the design of the mechanism - set initial design
        self.mechanism.factorizations[0].set_joint_connection_points_by_parameters(
            points_params[:len(self.mechanism.factorizations[0].dq_axes)])
        self.mechanism.factorizations[1].set_joint_connection_points_by_parameters(
            points_params[len(self.mechanism.factorizations[1].dq_axes):][::-1])

        if method == 'combinatorial_search':
            print("Starting combinatorial search algorithm...")
            cs = CombinatorialSearch(self.mechanism,
                                     linkage_length=result.fun,
                                     step_length=step_length,
                                     min_joint_segment_length=min_joint_segment_length,
                                     max_iters=max_iters)
            coll_free_points_params = cs.combinatorial_search(**kwargs)

        return coll_free_points_params


class CombinatorialSearch:
    """
    Combinatorial search algorithm of collision-free linkages.

    Algorithm by :footcite:t:`Li2020`.
    """
    def __init__(self,
                 mechanism: RationalMechanism,
                 linkage_length: float,
                 step_length: float = 10.0,  # TODO step length estimation
                 min_joint_segment_length: float = 0.001,
                 max_iters: int = 10):
        """
        Initialize the combinatorial search algorithm.

        :param RationalMechanism mechanism: The mechanism to optimize.
        :param float linkage_length: length of the linkage
        :param float step_length: length of the step, i.e. the shift distance value, see
            :ref:`combinatorial_search` for more detail
        :param float min_joint_segment_length: minimum length of the joint segment
        :param int max_iters: maximum number of iterations
        """
        self.mechanism = mechanism
        self.linkage_length = linkage_length
        self.step_length = step_length
        self.min_joint_segment_length = min_joint_segment_length
        self.max_iters = max_iters + 1

    def combinatorial_search(self, **kwargs):
        """
        Perform collision-free combinatorial search method.

        :return: list of collision-free points parameters
        :rtype: list
        """

        iter_start = kwargs.get('start_iteration', 1)
        iter_end = kwargs.get('end_iteration', self.max_iters)
        comb_links = kwargs.get('combinations_links', None)
        comb_joints = kwargs.get('combinations_joints', None)

        if comb_links is None:
            # check design for collisions
            init_collisions = self.mechanism.collision_check(only_links=True,
                                                             terminate_on_first=True)
        else:
            # skip initial collision check if combinations are provided
            init_collisions = []

        if init_collisions is not None:
            for i in range(iter_start, iter_end):
                coll_free_links_params = self.search_links(i, combinations=comb_links)

                if coll_free_links_params is not None:
                    print("")
                    print("Collision-free solution for links found, "
                          "starting joint search...")
                    coll_free_params = self.search_mechanism(coll_free_links_params,
                                                             combinations=comb_joints)

                    if coll_free_params is not None:
                        print("Search was successful, collision-free solution found.")
                        return coll_free_params
        else:
            print("Search was unsuccessful, collisions found.")
            return None

    def search_links(self, iteration: int, combinations: list = None):
        """
        Search for the solution of the combinatorial search algorithm, links only.

        Searches for the smallest polyline that is collision free (only links).

        :param iteration: iteration index
        :param list combinations: list of combinations to search links

        :return: list of collision-free points parameters
        :rtype: list
        """
        shift_val = iteration * self.linkage_length / self.step_length

        if combinations is None:
            combs = self._get_combinations_sequences(joints=False)
        else:
            combs = combinations

        for i, sequence in enumerate(combs):
            print("--- iteration: {}, shift_value: {}, sequence {} of {}: {}"
                  .format(iteration, shift_val, i + 1, len(combs), sequence))
            points_params = shift_val * np.asarray(sequence)
            points_params = [[param] for param in points_params]

            # update the design of the mechanism
            self.mechanism.factorizations[0].set_joint_connection_points_by_parameters(
                points_params[:len(self.mechanism.factorizations[0].dq_axes)])
            self.mechanism.factorizations[1].set_joint_connection_points_by_parameters(
                points_params[len(self.mechanism.factorizations[1].dq_axes):][::-1])

            colls = self.mechanism.collision_check(only_links=True,
                                                   terminate_on_first=True)

            if colls is None:
                return points_params

        print("No collision-free solution found for iteration: {}".format(iter))
        return None

    def search_mechanism(self, coll_free_links_params: list, combinations: list = None):
        """
        Search for the solution of the combinatorial search algorithm, including joints.

        Searches for the mechanism that is collision free (including joint segments).

        :param list coll_free_links_params: list of collision-free points parameters
        :param list combinations: list of combinations to search mechanism design

        :return: list of collision-free points parameters
        :rtype: list
        """
        shift_val = 0.5 * self.min_joint_segment_length

        if combinations is None:
            combs = self._get_combinations_sequences(joints=True)
        else:
            combs = combinations

        coll_free_links_params = [item * 2 for item in coll_free_links_params]

        for i, sequence in enumerate(combs):
            print("--- joint search. Shift_value: {}, sequence {} of {}: {}"
                  .format(shift_val, i + 1, len(combs), sequence))
            shift_seq = shift_val * np.asarray(sequence)

            points_params = [[params[0] + shift_seq[ii * 2],
                              params[1] + shift_seq[ii * 2 + 1]]
                             for ii, params in enumerate(coll_free_links_params)]

            # update the design of the mechanism
            self.mechanism.factorizations[0].set_joint_connection_points_by_parameters(
                points_params[:len(self.mechanism.factorizations[0].dq_axes)])
            self.mechanism.factorizations[1].set_joint_connection_points_by_parameters(
                points_params[len(self.mechanism.factorizations[1].dq_axes):][::-1])

            colls = self.mechanism.collision_check(only_links=False,
                                                   terminate_on_first=True)

            if colls is None:
                return points_params

        print("No collision-free solution found for iteration: {}".format(iter))
        return None

    def _get_combinations_sequences(self, joints: bool = False):
        """
        Get all combinations of the joint angles and shuffle them.

        :param bool joints: True if joints segments are searched for, False otherwise

        :return: list of all combinations of joint angles
        :rtype: list
        """
        if not joints:
            elements = [0, 1, -1]
            combs = list(product(elements, repeat=self.mechanism.num_joints))

            # remove the combination of all zeros, which was already tested
            if 0 in elements:
                combs.remove((0,) * self.mechanism.num_joints)

            return combs

        else:
            elements = [-1, 1]
            combs = list(product(elements, repeat=self.mechanism.num_joints * 2))

            # filter out the combs that have the same value for the 2 connection pts
            combs = [x for x in combs
                     if all(x[i] != x[i + 1] for i in range(0, len(x) - 1, 2))]

            # place the most promising combs first
            def generate_tuple(n):
                return tuple(-1 if i % 2 == 0 else 1 for i in range(n))

            tup = (generate_tuple(self.mechanism.num_joints)
                   + generate_tuple(self.mechanism.num_joints)[::-1])
            tup2 = (generate_tuple(self.mechanism.num_joints)[::-1]
                    + generate_tuple(self.mechanism.num_joints))

            combs.remove(tup)
            combs.insert(0, tup)
            combs.remove(tup2)
            combs.insert(0, tup2)

            return combs



from importlib.metadata import distribution


def is_package_installed(package_name: str) -> bool:
    """
    Check if a package is installed.
    """
    try:
        distribution(package_name)
        return True
    except ImportError:
        return False



from .AffineMetric import AffineMetric
from .CollisionFreeOptimization import CollisionFreeOptimization, CombinatorialSearch
from .DualQuaternion import DualQuaternion
from .DualQuaternionAction import DualQuaternionAction
from .ExudynAnalysis import ExudynAnalysis
from .FactorizationProvider import FactorizationProvider
from .Linkage import LineSegment, Linkage, PointsConnection
from .MiniBall import MiniBall
from .MotionFactorization import MotionFactorization
from .MotionInterpolation import MotionInterpolation
from .NormalizedLine import NormalizedLine
from .Plotter import Plotter
from .PointHomogeneous import PointHomogeneous
from .Quaternion import Quaternion
from .RationalBezier import RationalBezier, BezierSegment
from .RationalCurve import RationalCurve
from .RationalDualQuaternion import RationalDualQuaternion
from .RationalMechanism import RationalMechanism
from .TransfMatrix import TransfMatrix
from .CollisionAnalyser import CollisionAnalyser
from .StaticMechanism import StaticMechanism



from .RationalMechanism import RationalMechanism
from .RationalCurve import RationalCurve
from .RationalBezier import RationalBezier
from .MiniBall import MiniBall

import numpy


class CollisionAnalyser:
    def __init__(self):
        pass

    def check_two_objects(self, obj0, obj1):
        """
        Check if two objects collide.
        """
        obj0_type = self.get_object_type(obj0)
        obj1_type = self.get_object_type(obj1)

        if obj0_type == 'is_miniball' and obj1_type == 'is_miniball':
            return self.check_two_miniballs(obj0, obj1)

    @staticmethod
    def get_object_type(obj):
        """
        Get the type of an object.
        """
        if isinstance(obj, MiniBall):
            return 'is_miniball'

    @staticmethod
    def check_two_miniballs(ball0, ball1):
        """
        Check if two miniballs collide.
        """
        center_distance = numpy.linalg.norm(ball0.center.coordinates - ball1.center.coordinates)
        return center_distance < ball0.radius + ball1.radius



from .RationalMechanism import RationalMechanism
from .MotionFactorization import MotionFactorization
from .DualQuaternion import DualQuaternion
from .NormalizedLine import NormalizedLine


class StaticMechanism(RationalMechanism):
    """
    A class to represent a non-rational mechanism with a fixed number of joints

    This class is highly specialized and not intended for general use of Rational
    Linkages package. It can be used e.g. for obtaining the design (DH parameters, etc.)
    of a mechanism that has no rational parametrization.
    The joints  are assembled in a fixed loop-closure configuration. They are defined
    by a list of screw axes that are used to define the motion of the mechanism.

    :param list[NormalizedLine] screw_axes: A list of screw axes that define the
        kinematic structure of the mechanism.

    :ivar list[NormalizedLine] screws: A list of screw axes that define the kinematic
        structure of the mechanism.
    :ivar int num_joints: The number of joints in the mechanism.

    :example:

    .. testcode::

        # Define a 4-bar mechanism from points
        from rational_linkages import StaticMechanism


        l0 = NormalizedLine.from_two_points([0.0, 0.0, 0.0],
                                            [18.474, 30.280, 54.468])
        l1 = NormalizedLine.from_two_points([74.486, 0.0, 0.0],
                                            [104.321, 24.725, 52.188])
        l2 = NormalizedLine.from_two_points([124.616, 57.341, 16.561],
                                            [142.189, 91.439, 69.035])
        l3 = NormalizedLine.from_two_points([19.012, 32.278, 0.000],
                                            [26.852, 69.978, 52.367])

        m = StaticMechanism([l0, l1, l2, l3])

        m.get_design(unit='deg')

    """
    def __init__(self, screw_axes: list[NormalizedLine]):
        fake_factorization = [MotionFactorization([DualQuaternion()])]
        super().__init__(fake_factorization)

        self.screws = screw_axes
        self.num_joints = len(screw_axes)

        # redefine the factorization to use the screw axes
        self.factorizations[0].dq_axes = [DualQuaternion(axis.line2dq_array())
                                          for axis in screw_axes]

    def get_screw_axes(self) -> list[NormalizedLine]:
        """
        Overwrites the RationalMechanism method to return the screw axes of
        the mechanism.
        """
        return self.screws




from typing import Union

from .DualQuaternion import DualQuaternion
from .NormalizedLine import NormalizedLine
from .PointHomogeneous import PointHomogeneous


class DualQuaternionAction:
    """
    Strategy pattern class for acting on objects using Dual Quaternions

    So far, only acting on NormalizedLine and PointHomogeneous is implemented.
    """

    def __init__(self):
        pass

    def act(
        self,
        acting_object: Union[DualQuaternion, list[DualQuaternion]],
        affected_object: Union[NormalizedLine, PointHomogeneous],
    ) -> Union[NormalizedLine, PointHomogeneous]:
        """
        Act on an object using a Dual Quaternion

        :param acting_object: DualQuaternion or MotionFactorization
        :param affected_object: NormalizedLine or PointHomogeneous

        :return: NormalizedLine or PointHomogeneous

        :example:

        .. testcode::

            #  Act on line with a Dual Quaternion

            from rational_linkages import DualQuaternion, NormalizedLine

            dq = DualQuaternion([1, 2, 3, 4, 0.1, 0.2, 0.3, 0.4])
            line = NormalizedLine.from_direction_and_point([0, 0, 1], [0, -2, 0])

            line_after_half_turn = dq.act(line)

        """
        # check if the object is line or point
        affected_obj_type = self._analyze_affected_object(affected_object)

        # prepare the acting object (if it is a MotionFactorization, multiply)
        acting_obj = self._prepare_acting_object(acting_object)

        match affected_obj_type:
            case "is_line":
                return self._act_on_line(acting_obj, affected_object)
            case "is_point":
                return self._act_on_point(acting_obj, affected_object)

    @staticmethod
    def _analyze_affected_object(affected_object) -> str:
        """
        Analyze the affected object

        :param affected_object: NormalizedLine or PointHomogeneous

        :return: str - 'is_line' or 'is_point'
        """
        if isinstance(affected_object, NormalizedLine):
            return "is_line"
        elif isinstance(affected_object, PointHomogeneous):
            return "is_point"
        else:
            raise TypeError(
                "Other types than NormalizedLine or "
                "PointHomogeneous not yet implemented"
            )

    @staticmethod
    def _prepare_acting_object(
        acting_object: Union[DualQuaternion, list[DualQuaternion]]
    ) -> DualQuaternion:
        """
        Prepare the acting object

        :param acting_object: DualQuaternion or MotionFactorization

        :return: DualQuaternion
        """
        if isinstance(acting_object, DualQuaternion):
            return acting_object
        else:
            factors_multiplied = DualQuaternion()
            for factor in acting_object:
                factors_multiplied = factors_multiplied * factor
            return factors_multiplied

    @staticmethod
    def _act_on_line(
        acting_dq: DualQuaternion, affected_object: NormalizedLine
    ) -> NormalizedLine:
        """
        Act on a line with a Dual Quaternion

        :param acting_dq: DualQuaternion
        :param affected_object: NormalizedLine

        :return: NormalizedLine
        """
        line_as_dq = DualQuaternion(affected_object.line2dq_array())

        do_action = acting_dq * line_as_dq * acting_dq.conjugate()

        return NormalizedLine.from_dual_quaternion(do_action)

    @staticmethod
    def _act_on_point(
        acting_dq: DualQuaternion, affected_object: PointHomogeneous
    ) -> PointHomogeneous:
        """
        Act on a point with a Dual Quaternion

        :param acting_dq: DualQuaternion
        :param affected_object: PointHomogeneous

        :return: PointHomogeneous
        """
        point_as_dq = DualQuaternion(affected_object.point2dq_array())

        do_action = acting_dq.eps_conjugate() * point_as_dq * acting_dq.conjugate()

        return PointHomogeneous.from_dual_quaternion(do_action)



from .RationalMechanism import RationalMechanism
from .Linkage import LineSegment

import sympy


class SingularityAnalysis:
    """
    Singularity analysis algorithm of collision-free linkages by :footcite:t:`Li2020`.
    """
    def __init__(self):
        pass

    def check_singularity(self, mechanism: RationalMechanism):
        """
        Check for singularity in the mechanism.

        :param RationalMechanism mechanism: The mechanism to check for singularity
        """
        # check for singularity
        jacobian = self.get_jacobian(mechanism.segments)

        def get_submatrices(matrix):
            submatrices = []
            for row_to_remove in range(matrix.rows):
                for col_to_remove in range(matrix.cols):
                    # Create a submatrix by removing one row and one column
                    submatrix = matrix.minor_submatrix(row_to_remove, col_to_remove)
                    submatrices.append(submatrix)
            return submatrices

        def sum_of_squared_determinants(matrix):
            submatrices = get_submatrices(matrix)
            return sum(submatrix.det() ** 2 for submatrix in submatrices)

        sum_det = sum_of_squared_determinants(jacobian)

        return sum_det

    def get_jacobian(self, segments: list[LineSegment]):
        """
        Get the algebraic Jacobian matrix of the mechanism.

        :param list[LineSegment] segments: The line segments of the mechanism.
        """
        algebraic_plucker_coords = [joint.equation
                                    for joint in segments if joint.type == 'j']

        # normalization



        jacobian = sympy.Matrix.zeros(6, len(algebraic_plucker_coords))
        for i, plucker_line in enumerate(algebraic_plucker_coords):
            jacobian[:, i] = plucker_line.screw

        return jacobian



import numpy as np
from typing import Union

from .utils import is_package_installed
from .RationalMechanism import RationalMechanism


if not is_package_installed('exudyn'):
    print("Warning: The Exudyn package is not installed, since it is not "
          "a dependency. To use this functionality, please install it "
          "manually from PyPI using 'pip install exudyn'.")


class ExudynAnalysis:
    """
    Class for dynamics analysis using Exudyn package.

    The Exudyn packages is not listed in this project's requirements, please install
    it manually. More information can be found in :ref:`documentation <exudyn_info>`
    or the Exudyn homepage: https://github.com/jgerstmayr/EXUDYN
    """
    def __init__(self, gravity: Union[np.ndarray, list[float]] = np.array([0, 0, -9.81])):
        """
        Initialize ExudynAnalysis object.

        :param Union[np.ndarray, list[float]] gravity: XYZ gravity vector
        """
        self.gravity = gravity

    def get_exudyn_params(self,
                          mechanism: RationalMechanism,
                          link_radius: float = 0.1,
                          scale: float = 1.0) -> tuple:
        """
        Get parameters for Exudyn simulation.

        This method is used to get parameters for Exudyn - a multibody dynamics
        simulation package. The parameters are used to create a multibody system
        and simulate the mechanism's dynamics.

        The parameters are: links_pts (positions of links connection points),
        links_lengths, body_dim (dimensions of rigid bodies), links_masses_pts
        (positions of links' center of gravity), joint_axes (joints unit axes),
        relative_links_pts (links connection points relative to its center of gravity).

        :param RationalMechanism mechanism: RationalMechanism object
        :param float link_radius: width of links
        :param float scale: scale length factor for links dimensions

        :return: links_pts, links_lengths, body_dim, links_masses_pts, joint_axes,
            relative_links_pts
        :rtype: tuple
        """
        # get positions of links connection points
        links_pts = self._links_points(mechanism)

        if scale != 1.0:
            links_pts = [[scale * pt for pt in pts] for pts in links_pts]

        # get joint axes
        joint_axes = self._joints_axes(mechanism)

        # get links lengths
        links_lengths = self._links_lengths(links_pts)

        # get links center of gravity positions
        links_masses_pts = self._links_center_of_gravity(links_pts)

        # body dimensions
        body_dim = [[length, link_radius, link_radius] for length in links_lengths]

        # relative link points
        relative_links_pts = self._relative_links_points(links_pts, links_masses_pts)

        return (links_pts, links_lengths, body_dim, links_masses_pts,
                joint_axes, relative_links_pts)

    @staticmethod
    def _links_points(mechanism: RationalMechanism) -> list:
        """
        Get links connection points in default configuration.

        :param mechanism: RationalMechanism object

        :return: list of points on links
        :rtype: list
        """
        # get points sequence
        nearly_zero = np.finfo(float).eps
        points = (mechanism.factorizations[0].direct_kinematics(
                  nearly_zero, inverted_part=True)
                  + mechanism.factorizations[1].direct_kinematics(
                    nearly_zero, inverted_part=True)[::-1])

        # rearamge points, so the base link has the first 2 points
        points = points[-1:] + points[:-1]

        return list(zip(points[::2], points[1::2]))

    @staticmethod
    def _relative_links_points(links_points: list, centers_of_gravity: list) -> list:
        """
        Get links connection points in default configuration, relative to its center
        of gravity.

        :param list links_points: list of point pairs tuples
        :param list centers_of_gravity: list of links' center of gravity positions

        :return: list of points on links
        :rtype: list
        """
        return [(pts[0] - cog, pts[1] - cog)
                for pts, cog in zip(links_points, centers_of_gravity)]

    @staticmethod
    def _links_lengths(links_points: list) -> list:
        """
        Get links lengths.

        :param list links_points: list of point pairs tuples

        :return: list of links lengths
        :rtype: list
        """
        return [np.linalg.norm(pts[1] - pts[0]) for pts in links_points]

    @staticmethod
    def _links_center_of_gravity(links_points: list) -> list:
        """
        Get positions of links' center of gravity.

        :param list links_points: list of point pairs tuples

        :return: list of links' center of gravity positions
        :rtype: list
        """
        return [(pts[0] + pts[1]) / 2 for pts in links_points]

    @staticmethod
    def _joints_axes(mechanism: RationalMechanism) -> list:
        """
        Get joints unit axes.

        :param RationalMechanism mechanism: RationalMechanism object

        :return: list of joints axes
        :rtype: list
        """
        axes = []
        for axis in mechanism.factorizations[0].dq_axes:
            direction, moment = axis.dq2line_vectors()
            axes.append(direction)

        axes_branch2 = []
        for axis in mechanism.factorizations[1].dq_axes:
            direction, moment = axis.dq2line_vectors()
            axes_branch2.append(direction)

        return axes + axes_branch2[::-1]




from copy import deepcopy
from typing import Union

import numpy as np
import sympy as sp

from .PointHomogeneous import PointHomogeneous
from .DualQuaternion import DualQuaternion

MotionFactorization = "MotionFactorization"


class RationalCurve:
    """
    Class representing rational curves in n-dimensional space, where the first row is
    homogeneous coordinate equation.

    This class allows you to work with rational curves defined by parametric equations.

    :ivar coeffs: Coefficients of parametric equations of the curve.
    :ivar dimension: The dimension of the curve, excluding the homogeneous coordinate.
    :ivar degree: The degree of the curve.
    :ivar symbolic: Symbolic expressions for the parametric equations of the curve.
    :ivar set_of_polynomials: A set of polynomials representing the curve.
    :ivar symbolic_inversed: Symbolic expressions for the parametric equations of the
        inversed curve.
    :ivar set_of_polynomials_inversed: A set of polynomials representing the inversed
        curve.
    :ivar is_motion: True if the curve is a motion curve, False otherwise.

    :examples:

    .. testcode::

        # Limancon of Pascal -- from polynomial equations


        import sympy as sp
        from rational_linkages import RationalCurve

        a = 1
        b = 0.5
        t = sp.Symbol('t')
        eq0 = sp.Poly((1+t**2)**2, t)
        eq1 = sp.Poly(b*(1-t**2)*(1+t**2) + a*(1-t**2)**2, t)
        eq2 = sp.Poly(2*b*t*(1+t**2) + 2*a*t*(1-t**2), t)
        curve = RationalCurve([eq0, eq1, eq2, eq0])

    .. testcode::

        # From coefficients

        import sympy as sp
        from rational_linkages import RationalCurve


        curve = RationalCurve.from_coeffs(np.array([[1., 0., 2., 0., 1.], [0.5, 0., -2., 0., 1.5], [0., -1., 0., 3., 0.], [1., 0., 2., 0., 1.]]))
    """

    def __init__(self, polynomials: list[sp.Poly]):
        """
        Initializes a RationalCurve object with the provided coefficients.

        :param polynomials: list of polynomial equations of the curve
        """
        self.set_of_polynomials = polynomials

        self.dimension = len(self.set_of_polynomials) - 1
        # Get the degree of the curve
        self.degree = 1
        for i in range(len(polynomials)):
            self.degree = max(self.degree, self.set_of_polynomials[i].degree())

        self.coeffs = self.get_coeffs()
        self._symbolic = None

        self.coeffs_inversed = self.inverse_coeffs()
        self._symbolic_inversed = None
        self._set_of_polynomials_inversed = None

        # check if the curve is a motion curve
        self.is_motion = self.dimension == 7
        self.is_affine_motion = self.dimension == 12

    @property
    def symbolic(self):
        """
        Get the vector symbolic expressions of the curve

        :return: list of symbolic expressions
        :rtype: list
        """
        if self._symbolic is None:
            self._symbolic, _ = self.get_symbolic_expressions(self.coeffs)

        return self._symbolic

    @property
    def symbolic_inversed(self):
        """
        Get the vector symbolic expressions of the inversed curve

        :return: list of symbolic expressions
        :rtype: list
        """
        if self._symbolic_inversed is None:
            self._symbolic_inversed, self._set_of_polynomials_inversed \
                = self.get_symbolic_expressions(self.coeffs_inversed)

        return self._symbolic_inversed

    @property
    def set_of_polynomials_inversed(self):
        """
        Get the set of polynomials representing the inversed curve

        :return: list of symbolic expressions
        :rtype: list
        """
        if self._set_of_polynomials_inversed is None:
            self._symbolic_inversed, self._set_of_polynomials_inversed \
                = self.get_symbolic_expressions(self.coeffs_inversed)

        return self._set_of_polynomials_inversed

    @classmethod
    def from_coeffs(cls, coeffs: np.ndarray) -> "RationalCurve":
        """
        Construct rational curve from coefficients

        :coeffs: np.ndarray - coefficients of the curve

        :returns: RationalCurve object from coefficients
        :rtype: RationalCurve
        """
        _, polynomials = cls.get_symbolic_expressions(coeffs)
        return cls(polynomials)

    @staticmethod
    def get_symbolic_expressions(coeffs: np.ndarray) -> tuple[list, list[sp.Poly]]:
        """
        Add a symbolic variable to the matrix of coefficients that describes the curve

        :param np.ndarray coeffs: coefficients of the curve

        :return: tuple of symbolic expressions list and list of sympy polynomials
        :rtype: tuple[list, list[sp.Poly]]
        """
        symbolic_expressions = []
        polynomials = []
        t = sp.Symbol("t")

        dimension = len(coeffs) - 1

        for i in range(dimension + 1):
            # Extract coefficients from the current row of the coefficient matrix and
            # create symbolic expressions
            row_coefficients = reversed(coeffs[i, :])
            symbolic_row_coeffs = [
                coefficient * t**j for j, coefficient in enumerate(row_coefficients)
            ]
            symbolic_expressions.append(sum(symbolic_row_coeffs))
            polynomials.append(sp.Poly(symbolic_expressions[i], t))

        return symbolic_expressions, polynomials

    def get_coeffs(self) -> np.ndarray:
        """
        Get the coefficients of the symbolic polynomial equations

        :return: np.array of coefficients
        :rtype: np.ndarray
        """
        # Obtain the coefficients
        coeffs = np.zeros((self.dimension + 1, self.degree + 1))
        for i in range(self.dimension + 1):
            # to fill all coeffs, check if the degree of the equation is the same
            # as the curve
            if len(self.set_of_polynomials[i].all_coeffs()) == self.degree + 1:
                coeffs[i, :] = np.array(self.set_of_polynomials[i].all_coeffs())
            else:  # if the degree of the equation is lower than the curve, check
                # the difference
                if not (self.set_of_polynomials[i].all_coeffs() == [0.0]
                        or self.set_of_polynomials[i].all_coeffs() == [0]):
                    # if the equation is not zero, fill the coeffs
                    degree_of_eq = self.set_of_polynomials[i].degree()
                    coeffs[i, self.degree - degree_of_eq :] = np.array(
                        self.set_of_polynomials[i].all_coeffs()
                    )
        return coeffs

    def __repr__(self):
        return f"RationalCurve({self.symbolic})"

    def curve2bezier_control_points(self,
                                    reparametrization: bool = False
                                    ) -> list[PointHomogeneous]:
        """
        Convert a curve to a Bezier curve using the Bernstein polynomials

        :param bool reparametrization: if True, the curve is mapped to the [-1,1]

        :return: list of Bezier control points
        :rtype: list[PointHomogeneous]
        """
        t = sp.Symbol("t")

        # Get the symbolic variables in the form of x00, x01, ... based on degree
        # of curve and dimension of space
        points = [
            [sp.Symbol("x%d_%d" % (i, j)) for j in range(self.dimension + 1)]
            for i in range(self.degree + 1)
        ]
        points_flattened = [var for variables in points for var in variables]

        # Get the Bernstein polynomial equations and Bernstein basis
        expression_list = self.get_bernstein_polynomial_equations(t, reparametrization=reparametrization)
        bernstein_basis = [0] * (self.dimension + 1)
        for i in range(self.dimension + 1):
            for j in range(self.degree + 1):
                bernstein_basis[i] += expression_list[j] * points[j][i]

        # Get the coefficients of the equations
        equations_coeffs = [
            sp.Poly((bernstein_basis[i] - self.symbolic[i]), t).all_coeffs()
            for i in range(self.dimension + 1)
        ]
        # Flatten the list
        equations_coeffs = [coeff for coeffs in equations_coeffs for coeff in coeffs]

        # Solve the equations
        points_sol = sp.linsolve(equations_coeffs, points_flattened)
        # Convert the solutions to numpy arrays (get points)
        points_array = np.array(points_sol.args[0], dtype="float64").reshape(
            self.degree + 1, self.dimension + 1)

        points_objects = [PointHomogeneous()] * (self.degree + 1)
        for i in range(self.degree + 1):
            points_objects[i] = PointHomogeneous(points_array[i, :])

        return points_objects

    def get_bernstein_polynomial_equations(
        self, t_var: sp.Symbol, reparametrization: bool = False, degree: int = None
    ) -> list:
        """
        Generate the Bernstein polynomial equation

        :param sp.Symbol t_var: symbolic variable
        :param bool reparametrization: if True, the curve is mapped to the [-1,1]
        :param degree: int - degree of the polynomial, if None (not specified),
            the degree of the curve is used

        :return: list of symbolic expressions
        :rtype: list
        """
        if degree is None:
            degree = self.degree

        if not reparametrization:
            t = t_var
        else:
            # Mapping of t to the interval [-1, 1]
            t = (t_var + 1) / 2

            # NOT WORKING because sp.Poly cannot handle
            # t = 1/sp.tan(t_var/2)
            # t = 1/t_var

        # Initialize the polynomial expression list
        expr = []

        # Generate the polynomial expression using the Bernstein polynomials
        for i in range(degree + 1):
            polynomial_expr = sp.binomial(degree, i) * t**i * (1 - t) ** (degree - i)
            #expr.append(sp.simplify(polynomial_expr))
            expr.append(polynomial_expr)
            # TODO: simplify bottleneck

        return expr

    def inverse_coeffs(self) -> np.ndarray:
        """
        Get the coefficients of the inverse curve

        :return: np.array of inversed coefficients
        :rtype: np.ndarray
        """
        inverse_coeffs = np.zeros((self.dimension + 1, self.degree + 1))
        for i in range(self.dimension + 1):
            inverse_coeffs[i, :] = self.coeffs[i, :][::-1]

        return inverse_coeffs

    def inverse_curve(self) -> "RationalCurve":
        """
        Get the inverse curve

        :return: inversed rational curve
        :rtype: RationalCurve
        """
        return RationalCurve.from_coeffs(self.inverse_coeffs())

    def curve(self) -> "RationalCurve":
        """
        Get the rational curve (itself) - suitable for subclasses, returns the
        superclass object

        :return: RationalCurve
        :rtype: RationalCurve
        """
        return RationalCurve(self.set_of_polynomials)

    def extract_expressions(self) -> list:
        """
        Extract the expressions of the curve

        :return: list of expressions of the curve (avoiding sp.Poly class)
        :rtype: list
        """
        return [self.set_of_polynomials[i].expr
                for i in range(len(self.set_of_polynomials))]

    def evaluate(self, t_param: Union[float, np.ndarray],
                 inverted_part: bool = False) -> np.ndarray:
        """
        Evaluate the curve for given t and return in the form of dual quaternion vector

        :param float, np.ndarray t_param: parameter of the motion curve
        :param bool inverted_part: if True, return the inverted part of the curve

        :return: pose of the curve as a dual quaternion vector
        :rtype: np.ndarray
        """
        t = sp.Symbol("t")
        if inverted_part:
            return np.array(
                [
                    self.set_of_polynomials_inversed[i].subs(t, t_param).evalf()
                    for i in range(len(self.set_of_polynomials_inversed))
                ],
                dtype="float64",
            )
        else:
            return np.array(
                [
                    self.set_of_polynomials[i].subs(t, t_param).evalf()
                    for i in range(len(self.set_of_polynomials))
                ],
                dtype="float64",
            )

    def evaluate_as_matrix(self, t_param, inverted_part: bool = False) -> np.ndarray:
        """
        Evaluate the curve for given t and return in the form of a transformation matrix

        :param float t_param: parameter of the motion curve
        :param bool inverted_part: if True, return the inverted part of the curve

        :return: pose of the curve as a matrix
        :rtype: np.ndarray
        """
        from .DualQuaternion import DualQuaternion

        dq = DualQuaternion(self.evaluate(t_param, inverted_part))
        return dq.dq2matrix()

    def factorize(self, use_rationals: bool = False) -> list[MotionFactorization]:
        """
        Factorize the curve into motion factorizations

        :param bool use_rationals: if True, force the factorization in QQ to return
            rational numbers

        :return: list of MotionFactorization objects
        :rtype: list[MotionFactorization]
        """
        if type(self) != RationalCurve:
            raise TypeError("Can factorize only for a rational curve or motion "
                            "factorization")

        from .FactorizationProvider import FactorizationProvider

        factorization_provider = FactorizationProvider(use_rationals=use_rationals)
        return factorization_provider.factorize_motion_curve(self)

    def get_plot_data(self, interval: Union[str, tuple] = (0, 1), steps: int = 50) -> (
            tuple)[np.ndarray, np.ndarray, np.ndarray]:
        """
        Get the data to plot the curve in 3D

        :param Union[str, tuple] interval: interval of the parameter t, if 'closed',
            the closed-loop curve is parametrized using tangent half-angle substitution
        :param int steps: number of numerical steps in the interval

        :return: tuple of np.ndarray - (x, y, z) coordinates of the curve
        :rtype: tuple[np.ndarray, np.ndarray, np.ndarray]
        """
        t = sp.Symbol("t")

        if interval == 'closed':
            # tangent half-angle substitution for closed curves
            t_space = np.tan(np.linspace(-np.pi/2, np.pi/2, steps + 1))
        else:
            t_space = np.linspace(interval[0], interval[1], steps)

        # make a copy of the polynomials and append the homogeneous coordinate
        # to the Z-equation place in the list if in 2D, so later z = 1
        polynoms = deepcopy(self.set_of_polynomials)
        if self.dimension == 2:
            polynoms.append(sp.Poly(polynoms[0], t))

        # plot the curve
        curve_points = [PointHomogeneous()] * steps
        for i in range(steps):
            point = self.evaluate(t_space[i])

            # if it is a pose in SE3, convert it to a point via matrix mapping
            if self.is_motion:
                point = DualQuaternion(point).dq2point_via_matrix()
                point = np.concatenate((np.array([1]), point))
            elif self.is_affine_motion:
                point = point[:4]

            curve_points[i] = PointHomogeneous([point[0], point[-3], point[-2], point[-1]])
        x, y, z = zip(*[curve_points[i].normalized_in_3d() for i in range(steps)])
        return x, y, z

    def get_curve_in_pr12(self) -> "RationalCurve":
        """
        Get the representation of the curve in PR12

        :return: curve in PR12
        :rtype: RationalCurve
        """
        if not self.is_motion:
            raise ValueError("The curve is not a motion curve, cannot convert to PR12")

        t = sp.Symbol("t")

        # convert the motion curve to a dual quaternion, then map to matrix
        curve_matrix = DualQuaternion(self.symbolic).dq2matrix(normalize=False)

        # save the not normalized coordinate
        curve_p = curve_matrix[0, 0]
        # transpose the matrix so the flatten() provides right order (vector by vector)
        curve_r12 = curve_matrix[1:4, :].T.flatten()

        # create PR12 vector of polynomial equations
        curve_pr12 = np.concatenate((np.array([curve_p]), curve_r12))
        curve_poly = [sp.Poly(curve, t) for curve in curve_pr12]

        return RationalCurve(curve_poly)

    def split_in_beziers(self,
                         metric: Union[str, "AffineMetric"] = "euclidean",
                         min_splits: int = 0) -> list["BezierSegment"]:
        """
        Split the curve into Bezier curves with positive weights of control points.

        The curve is split into Bezier curves using the De Casteljau algorithm.

        :param int min_splits: minimal number of splits to be performed
        :param Union[str, AffineMetric] metric: metric for the optimization

        :return: list of RationalBezier objects
        :rtype: list[RationalBezier]
        """
        if self.is_motion:
            curve = self.get_curve_in_pr12()
        else:
            raise ValueError("The curve is not a motion curve, cannot "
                             "split into Bezier curves")

        from .RationalBezier import BezierSegment  # method import

        # obtain Bezier curves for the curve and its reparametrized inverse part
        bezier_curve_segments = [
            # reparametrize the curve from the intervals [-1, 1]
            BezierSegment(curve.curve2bezier_control_points(reparametrization=True),
                          metric=metric,
                          t_param=(False, [-1.0, 1.0])),
            BezierSegment(curve.inverse_curve().curve2bezier_control_points(
                reparametrization=True),
                          metric=metric,
                          t_param=(True, [-1.0, 1.0]))
        ]

        # split the Bezier curves until all control points have positive weights
        while True:
            new_segments = [
                part for b_curve in bezier_curve_segments
                for part in (
                    b_curve.split_de_casteljau(metric=metric) if b_curve.check_for_control_points_at_infinity() or b_curve.check_for_negative_weights() else [b_curve])
            ]

            if not any(
                    b_curve.check_for_control_points_at_infinity() or b_curve.check_for_negative_weights()
                    for b_curve in new_segments):
                if len(new_segments) < min_splits:
                    new_segments = [
                        part for b_curve in new_segments
                        for part in b_curve.split_de_casteljau(metric=metric)
                    ]
                else:
                    bezier_curve_segments = new_segments
                    break

            bezier_curve_segments = new_segments

        return bezier_curve_segments

    def get_path_length(self, num_of_points: int = 100) -> float:
        """
        Get the length of the curve path

        Evaluates the curve in the given number of points and sums the distances between.

        :param int num_of_points: number of discrete points to evaluate the curve

        :return: length of the curve path
        :rtype: float
        """
        t_space = np.tan(np.linspace(-np.pi/2, np.pi/2, num_of_points))
        poses = [self.evaluate(t) for t in t_space]

        points = [DualQuaternion(p).dq2point_via_matrix()
                  for p in poses]

        return np.sum(np.linalg.norm(np.diff(points, axis=0), axis=1))


